Attribute VB_Name = "bfit"
Option Explicit

Public Const Pi = 3.14159
Public Const C0 = 0.3989423        'Used in Gaussian equation, equals 1/SQR(2*Pi)
Public Const LamdaD = 0.000000000155125 'Total decay constant for 238U (yr^-1)
Public Const GF = 0.5              'Geometry factor
Public Const KernelFactor = 0.6    'Scaling factor used in PD plots
Public PkWidthFactor As Single     'Scaling factor relating SE(Z) to peak width

Public progname As String
Public Num As Integer
Public PkNum As Integer

Public Ns(1 To maxGrains) As Integer
Public Ni(1 To maxGrains) As Integer
Public Ni0(1 To maxGrains) As Single
Public Nt(1 To maxGrains) As Integer

Public grainAge(1 To maxGrains) As Single
Public Zgrain(1 To maxGrains) As Single
Public Zerr(1 To maxGrains) As Single

Public PkAge(1 To maxPeaks) As Single
Public SEPkAge(1 To maxPeaks, 4) As Single
Public PkTheta(1 To maxPeaks) As Single
Public PkFrac(1 To maxPeaks) As Single
Public SEPkFrac(1 To maxPeaks) As Single
Public PkSDz(1 To maxPeaks) As Single

Public RERhoD0 As Single
Public REZeta0 As Single
Public Zeta0 As Single
Public RhoD0 As Single
Public Uglass0 As Single
Public SqSize As Single
Public Iter As Integer

Public AgeMin As Single
Public AgeMax As Single
Public QMeanZerr
Public EstPkWidth

Public file As String
Public title As String

Public Function EM(PkTheta() As Single, PkFrac() As Single) As Single
  '============================================================================
  '... Uses the EM method outlined in Galbraith (1988) to find the best-fit
  ' peaks (see Titterington and others, 1985, p. 84-97 for further details).
  ' Convergence is defined as when the relative changes in the Log-Likelihood
  ' value are less than the constant Rtol.
  '============================================================================
  Const Rtol = 0.00001
  Dim i As Integer
  Dim U As Integer
  Dim SumFiu As Double
  Dim SumPkfrac As Single
  Dim Sum1 As Single, Sum2 As Single
  Dim ChiSq
  Dim Rchange
  Dim Itest1 As Integer
  Dim p(1 To maxPeaks, 1 To maxGrains) As Single
  Dim SmallLogLike As Single
  Dim OldLogLike As Single
  Dim LogLike As Single

  OldLogLike = 0
  Iter = 0
  Do
    Iter = Iter + 1
    '... Calculate new values for the array P() and LogLikelihood parameter
    LogLike = 0
    For U = 1 To Num
      SumFiu = 0
      For i = 1 To PkNum
        p(i, U) = PkFrac(i) * Fiu(Ns(U), Ni(U), Nt(U), PkTheta(i))
        SumFiu = SumFiu + p(i, U)
      Next i
      For i = 1 To PkNum
        p(i, U) = p(i, U) / SumFiu
      Next i
      LogLike = LogLike + log(SumFiu)
    Next U
    '... Update proportions of the peaks in Pkfrac()
    For i = 1 To PkNum
      SumPkfrac = 0
      For U = 1 To Num
        SumPkfrac = SumPkfrac + p(i, U)
      Next U
      PkFrac(i) = SumPkfrac / Num
    Next i
    '... Update peak theta ratios in PkTheta()
    For i = 1 To PkNum
      Sum1 = 0
      Sum2 = 0
      For U = 1 To Num
        Sum1 = Sum1 + p(i, U) * Ns(U)
        Sum2 = Sum2 + p(i, U) * Nt(U)
      Next U
      PkTheta(i) = Sum1 / Sum2
    Next i
    ChiSq = Chi2()
    
    '... Check for convergence: the relative size of the final steps must
    ' remain small, less than Rtol. The routine now includes a limit
    ' on the maximum number of iterations that can occur without advancing
    ' the solution.
    Rchange = Abs(0.5 * (LogLike - OldLogLike) / (LogLike + OldLogLike))
    '!!! Print USING; "####)      ######.######  ##.#####^^^^"; Iter%; LogLike; Rchange
    If Rchange <= Rtol Then
      Itest1 = Itest1 + 1
      If Itest1 = 3 Then Exit Do
    Else
      Itest1 = 0
    End If
    OldLogLike = LogLike

    If Iter > 300 Then    '!!!
      MsgBox "WARNING: LOCAL MINIMUM POORLY DEFINED" + vbCrLf + "CHECK THAT BEST-FIT SOLUTION IS REASONABLE"
      Exit Do
    End If
  Loop
  EM = LogLike
End Function

Private Function FactLn(n As Integer) As Single
  '============================================================================
  '... Returns Ln[N!], which is the natural logrithm of N factorial.
  ' From Press and others (1986, p. 158).
  '============================================================================
  '... On first call, initialize table of values to -1. Flag% is used to
  ' test for first call condition (Flag% has an initial value of False
  ' at start of program).
  
  Static Flag As Integer
  Static a(1 To 100) As Single
  Dim i As Integer

  If Not Flag Then
    For i = 1 To 100
      a(i) = -1
    Next i
    Flag = True
  End If
  '... Test for negative values
  If n < 0 Then  '!!!
    MsgBox "Negative factorials are not permitted"
    Stop
  End If
  '... Get FactLn result
  If n <= 99 Then
    '... Put value in table if not already there
    If a(n + 1) < 0 Then a(n + 1) = GammLn(n + 1)
    FactLn = a(n + 1)
  Else
    '... Out of range of lookup table
    FactLn = GammLn(n + 1)
  End If
End Function

Public Function GammLn(xx As Single) As Single
  '... Calculates the logrithm of the gamma function using the method of
  ' Lanczos. See Press and others (1986, p. 157) for details.
  Const COF1 = 76.18009173
  Const COF2 = -86.50532033
  Const COF3 = 24.01409822
  Const COF4 = -1.231739516
  Const COF5 = 0.00120858003
  Const COF6 = -0.00000536382
  Const STP = 2.50662827465
  Const HALF = 0.5
  Const ONE = 1#
  Const FPF = 5.5
  Dim x As Double, tmp As Double, SER As Double

  x = xx - ONE
  tmp = x + FPF
  tmp = (x + HALF) * log(tmp) - tmp
  x = x + ONE
  SER = ONE + COF1 / x
  x = x + ONE
  SER = SER + COF2 / x
  x = x + ONE
  SER = SER + COF3 / x
  x = x + ONE
  SER = SER + COF4 / x
  x = x + ONE
  SER = SER + COF5 / x
  x = x + ONE
  SER = SER + COF6 / x
  GammLn = tmp + log(STP * SER)
End Function

Function Chi2() As Single
  '============================================================================
  '... Calculates Chi^2 parameter for the binomial model of Galbraith
  ' and Green. PkTheta() contains the theta ratios for the best-fit peaks, and
  ' Pkfrac() contains the proportions of the total distribution for the peaks.
  ' Num% (shared) = number of dated grains
  ' PkNum% (shared) = number of peaks to be found
  ' Ns() (shared) = number of spontaneous tracks counted per grain
  ' Nt() (shared) = total number of tracks counted per grain
  ' Chi2 = function is returned with Chi^2 value
  '============================================================================
  Dim Mu As Single, ss As Single
  Dim i As Integer
  Dim Factor1 As Single, Factor2 As Single
  Dim C2 As Single, dev As Single, Var As Single
  Dim nnt As Single

  '... Find mean and sum of squares for solution vector
  Mu = 0
  ss = 0
  For i = 1 To PkNum
    Mu = Mu + PkFrac(i) * PkTheta(i)
    ss = ss + PkFrac(i) * PkTheta(i) * PkTheta(i)
  Next i

  '... Find Chi2 misfit for this particular solution vector
  Factor1 = Mu * (1 - Mu)
  Factor2 = ss - Mu * Mu
  C2 = 0
  For i = 1 To Num
    nnt = Nt(i)
    '... if Nt(i) = 0, then Ns(i)=0, so contribution to Chi2 is zero.
    If nnt > 0 Then
      dev = Ns(i) - Mu * nnt
      Var = Factor1 * nnt + nnt * (nnt - 1) * Factor2
      C2 = C2 + dev * dev / Var
    End If
  Next i
  Chi2 = C2
End Function

Function Fiu(Ru As Integer, Su As Integer, Tu As Integer, PkTheta As Single) As Single
  '============================================================================
  '... Evaluates the contribution that the Uth grain makes to the binomial
  ' probability density for the Ith peak. The binomial probability density
  ' calculation halts with an overflow error for Ru% greater than about 500.
  ' As a result, the Gaussian probability density distribution is used to
  ' approximate the binomial probability density function for Ru% > 100 and
  ' PKtheta between 5% and 95% (see Taylor, 1982, p. 194, for details).
  ' Ru%, Su%, Tu% = spontaneous, induced, and total track for the Uth grain.
  ' PkTheta = mean theta ratio for the Ith peak.
  ' The constant C0 = 1/SQR(2*Pi) and is set in the main program.
  ' ... 9/15/04: added statement to ensure a minimum value for Fiu, in order
  ' to avoid problems with round-off errors.
  ' ... 6/27/2007: Completed full calculation of binomial probability mass in
  ' log units to avoid overflow errors.
  '============================================================================
  Dim Mean As Single, SD As Single, Z As Single
  Dim tmp As Double, lnbico As Double, lny As Double
  If Tu > 100 And PkTheta > 0.05 And PkTheta < 0.95 Then
    Mean = Tu * PkTheta
    SD = Sqr(Mean * (1 - PkTheta))
    Z = (Ru - Mean) / SD
    tmp = (C0 / SD) * Exp(-Z * Z / 2)
  Else
    lnbico = FactLn(Tu) - FactLn(Ru) - FactLn(Tu - Ru)
    lny = lnbico + Ru * log(PkTheta) + Su * log(1 - PkTheta)
    tmp = Exp(lny)
  End If
  If tmp <= 0.0000000001 Then
    Fiu = 0.0000000001
  Else
    Fiu = tmp
  End If
End Function

Public Function StdErr() As Single
  '============================================================================
  '... Calculates peak ages and approximate standard errors for the best-fit
  ' parameters using the method outlined in Galbraith (1988, p. 280). Note that
  ' the method he uses produces a scaled covariance matrix.
  ' ZErr() (shared) = SE(Z) for grains
  '... 9/15/2004: fixed division by zero when SumFiu = 0
  '============================================================================
  Dim p(1 To maxPeaks, 1 To maxGrains) As Single
  Dim a(1 To maxPeaks - 1, 1 To maxPeaks - 1) As Single
  Dim b(1 To maxPeaks - 1, 1 To maxPeaks) As Single
  Dim c(1 To maxPeaks, 1 To maxPeaks) As Single
  Dim Covar(1 To 2 * maxPeaks - 1, 1 To 2 * maxPeaks - 1) As Single
  Dim W(1 To 2 * maxPeaks - 1) As Single
  Dim V(1 To 2 * maxPeaks - 1, 1 To 2 * maxPeaks - 1) As Single
  Dim i As Integer, U As Integer, j As Integer

  Dim cNum As Single

  '... Calculate the array P()
  Dim SumFiu As Double

  For U = 1 To Num
    SumFiu = 0
    For i = 1 To PkNum
      p(i, U) = PkFrac(i) * Fiu(Ns(U), Ni(U), Nt(U), PkTheta(i))
      SumFiu = SumFiu + p(i, U)
    Next i
    For i = 1 To PkNum
        '... Check for SumFiu = 0, which implies that all p(i,U) =0
        p(i, U) = p(i, U) / SumFiu
    Next i
  Next U

  '... Construct A(), the upper right quadrant of the Inverse Covariance matrix
  Dim PIk As Single
  Dim Pku As Single
  Dim Term As Single

  PIk = PkFrac(PkNum)
  For i = 1 To PkNum - 1
    For j = 1 To PkNum - 1
      a(i, j) = 0
      For U = 1 To Num
        Pku = p(PkNum, U)
        Term = (p(i, U) / PkFrac(i) - Pku / PIk) * (p(j, U) / PkFrac(j) - Pku / PIk)
        a(i, j) = a(i, j) + Term
      Next U
    Next j
  Next i

  '... Construct B(), the diagonal quadrants of the Inverse Covariance matrix
  Dim del1 As Integer, del2 As Integer
  Dim Aju As Single

  For i = 1 To PkNum% - 1
    For j = 1 To PkNum
      b(i, j) = 0
      For U = 1 To Num
        Pku = p(PkNum, U)
        If i = j Then del1 = 1 Else del1 = 0
        If PkNum = j Then del2 = 1 Else del2 = 0
        Term = p(i, U) / PkFrac(i) - Pku / PIk - del1 / PkFrac(i) + del2 / PIk
        Aju = Ns(U) - PkTheta(j) * Nt(U)
        b(i, j) = b(i, j) + p(j, U) * Aju * Term
      Next U
    Next j
  Next i

  '... Construct C(), the lower left quadrant of the Inverse Covariance matrix
  Dim Aiu As Single

  For i = 1 To PkNum
    For j = 1 To PkNum
      c(i, j) = 0
      For U = 1 To Num
        '... Calculate first derivatives of log(fiu) wrt beta(i)
        ' and log(fju) wrt beta(j)
        Aiu = Ns(U) - PkTheta(i) * Nt(U)
        Aju = Ns(U) - PkTheta(j) * Nt(U)
        Term = p(j, U) * Aiu * Aju
        If i = j Then
          '... Add second derivative of log(fiu) wrt beta(i)
          Term = Term - (Aiu * Aiu - PkTheta(i) * (1 - PkTheta(i)) * Nt(U))
        End If
        c(i, j) = c(i, j) + p(i, U) * Term
      Next U
    Next j
  Next i

  '... Assembly inverse of covariance matrix
  For i = 1 To PkNum - 1
    For j = 1 To PkNum - 1
      Covar(i, j) = a(i, j)
    Next j
  Next i

  For i = 1 To PkNum - 1
    For j = 1 To PkNum
      Covar(i, j + PkNum - 1) = b(i, j)
    Next j
  Next i

  For i = 1 To PkNum
    For j = 1 To PkNum - 1
      Covar(i + PkNum - 1, j) = b(j, i)
    Next j
  Next i

  For i = 1 To PkNum
    For j = 1 To PkNum
      Covar(i + PkNum - 1, j + PkNum - 1) = c(i, j)
    Next j
  Next i

  '... Invert the matrix Covar using the SVD routine
  Dim M As Integer

  M = 2 * PkNum - 1
  Call SvdCmp(Covar(), W(), V(), M, M)
  Call SvdInv(Covar(), W(), V(), cNum, M)
  '!!! Print "Condition number for inversion of covariance matrix: "; Cnum

  '... Calculate standard errors for the fraction parameter for 1 to PkNum%-1
  For i = 1 To PkNum - 1
    SEPkFrac(i) = Sqr(Abs(Covar(i, i)))
  Next i

  '... Calculate standard error for the last fraction parameter (see Press
  ' et al., 1992, p. 692, for details about this calculation)
  Dim Sum As Single

  Sum = 0
  For i = 1 To PkNum - 1
    For j = 1 To PkNum - 1
      Sum = Sum + Covar(i, j)
    Next j
  Next i
  SEPkFrac(PkNum) = Sqr(Abs(Sum))

  '... Calculate peak ages, standard errors, and estimated peak widths
  Dim BetaPk As Single
  Dim SE As Single
  Dim SumNt As Single

  For i = 1 To PkNum
    BetaPk = log(PkTheta(i) / (1 - PkTheta(i)))
    PkAge(i) = TaufromBeta(BetaPk)
    j = i + PkNum - 1
    SE = Sqr(Abs(Covar(j, j)) + RERhoD0 * RERhoD0 + REZeta0 * REZeta0)
    SEPkAge(i, 1) = TaufromBeta(BetaPk - SE) - PkAge(i)
    SEPkAge(i, 2) = TaufromBeta(BetaPk + SE) - PkAge(i)
    SEPkAge(i, 3) = TaufromBeta(BetaPk - 1.96 * SE) - PkAge(i)
    SEPkAge(i, 4) = TaufromBeta(BetaPk + 1.96 * SE) - PkAge(i)
    SumNt = 0
    SumFiu = 0
    For U = 1 To Num
      SumNt = SumNt + p(i, U) * Nt(U)
      SumFiu = SumFiu + p(i, U)
    Next U
    SumNt = SumNt / SumFiu
    PkSDz(i) = Sqr(1 / (SumNt * PkTheta(i) * (1 - PkTheta(i))))
  Next i

  '... Reclaim temporary array space
  Erase p, a, b, c, Covar, W, V

  StdErr = cNum
End Function

Sub SvdCmp(a() As Single, W() As Single, V() As Single, M As Integer, n As Integer)
  '============================================================================
  '... Calculates the singular value decomposition for the matrix A().
  ' The routine is from Press and others (1986, p. 60 to 64).
  ' On return, A() is converted into the SVD matrix U(). The other SVD matrices
  ' are contained in W() and V().
  ' Dimensions of A(), W(), and V() should be set in the calling procedure,
  ' as such: Dim A(M%, N%), W(N%), V(N%, N%)
  ' 11/18/91:   Fixed a bug related to the SGN() function and got rid of all
  '     GOTO statements.
  '============================================================================
  Dim RV1() As Single
  ReDim RV1(1 To n)
  Dim G As Single
  Dim S As Single
  Dim f As Single
  Dim H As Single
  Dim Scal As Single
  Dim Anorm As Single
  Dim Dum As Single
  Dim i As Integer, l As Integer, k As Integer, j As Integer

  '... Householder reduction to bidiagonal form
  G = 0!
  Scal = 0!
  Anorm = 0!
  For i = 1 To n
    l = i + 1
    RV1(i) = Scal * G
    G = 0!
    S = 0!
    Scal = 0!
    If i <= M Then
      For k = i To M
        Scal = Scal + Abs(a(k, i))
      Next k
      If Scal <> 0! Then
        For k = i To M
          a(k, i) = a(k, i) / Scal
          S = S + a(k, i) * a(k, i)
        Next k
        f = a(i, i)
        G = Sqr(S)
        If f >= 0 Then G = -G
        H = f * G - S
        a(i, i) = f - G
        If i <> n Then
          For j = l To n
            S = 0!
            For k = i To M
              S = S + a(k, i) * a(k, j)
            Next k
            f = S / H
            For k = i To M
              a(k, j) = a(k, j) + f * a(k, i)
            Next k
          Next j
        End If
        For k = i To M
          a(k, i) = Scal * a(k, i)
        Next k
      End If
    End If
    W(i) = Scal * G
    G = 0!
    S = 0!
    Scal = 0!
    If (i <= M) And (i <> n) Then
      For k = l To n
        Scal = Scal + Abs(a(i, k))
      Next k
      If Scal <> 0! Then
        For k = l To n
          a(i, k) = a(i, k) / Scal
          S = S + a(i, k) * a(i, k)
        Next k
        f = a(i, l)
        G = Sqr(S)
        If f >= 0 Then G = -G
        H = f * G - S
        a(i, l) = f - G
        For k = l To n
          RV1(k) = a(i, k) / H
        Next k
        If i <> M Then
          For j = l To M
            S = 0!
            For k = l To n
              S = S + a(j, k) * a(i, k)
            Next k
            For k = l To n
              a(j, k) = a(j, k) + S * RV1(k)
            Next k
          Next j
        End If
        For k = l To n
          a(i, k) = Scal * a(i, k)
        Next k
      End If
    End If
    Dum = Abs(W(i)) + Abs(RV1(i))
    If Dum > Anorm Then
      Anorm = Dum
    End If
  Next i

  '... accumulation of right-hand transformations
  For i = n To 1 Step -1
    If i < n Then
      If G <> 0! Then
        For j = l To n
          '... double division to avoid possible underflow
          V(j, i) = (a(i, j) / a(i, l)) / G
        Next j
        For j = l To n
          S = 0!
          For k = l To n
            S = S + a(i, k) * V(k, j)
          Next k
          For k = l To n
            V(k, j) = V(k, j) + S * V(k, i)
          Next k
        Next j
      End If
      For j = l To n
        V(i, j) = 0!
        V(j, i) = 0!
      Next j
    End If
    V(i, i) = 1!
    G = RV1(i)
    l = i
  Next i

  '... accumulation of left-hand transformations
  For i = n To 1 Step -1
    l = i + 1
    G = W(i)
    If i < n Then
      For j = l To n
        a(i, j) = 0!
      Next j
    End If
    If G <> 0! Then
      G = 1! / G
      If i <> n Then
        For j = l To n
          S = 0!
          For k = l To M
            S = S + a(k, i) * a(k, j)
          Next k
          f = (S / a(i, i)) * G
          For k = i To M
            a(k, j) = a(k, j) + f * a(k, i)
          Next k
        Next j
      End If
      For j = i To M
        a(j, i) = a(j, i) * G
      Next j
    Else
      For j = i To M
        a(j, i) = 0!
      Next j
    End If
    a(i, i) = a(i, i) + 1!
  Next i

  '... Diagonalization of the bidiagonal form
  Dim ITS As Integer
  Dim NM As Integer
  Dim c As Single
  Dim Z As Single
  Dim x As Single
  Dim y As Single
  Dim GsgnF As Single

  For k = n To 1 Step -1
    '... loop for each singular value
    For ITS = 1 To 30
      '... loop until convergence or exceeds maximum iterations
      For l = k To 1 Step -1
        '... test for splitting; note that RV1(1) is always zero
        NM = l - 1
        If (Abs(RV1(l)) + Anorm) = Anorm Then Exit For
        If (Abs(W(NM)) + Anorm) = Anorm Then Exit For
      Next l
      If (Abs(RV1(l)) + Anorm) <> Anorm Then
        '... cancellation of RV1(L) if L > 1
        c = 0!
        S = 1!
        For i = l To k
          f = S * RV1(i)
          If (Abs(f) + Anorm) <> Anorm Then
            G = W(i)
            H = Sqr(f * f + G * G)
            W(i) = H
            H = 1! / H
            c = (G * H)
            S = -(f * H)
            For j = 1 To M
              y = a(j, NM)
              Z = a(j, i)
              a(j, NM) = (y * c) + (Z * S)
              a(j, i) = -(y * S) + (Z * c)
            Next j
          End If
        Next i
      End If
      Z = W(k)
      If l = k Then
        '... routine has converged
        If Z < 0! Then
          '... singular value is made nonnegative
          W(k) = -Z
          For j = 1 To n
            V(j, k) = -V(j, k)
          Next j
        End If
        Exit For
      End If
      '... exceeded number of iterations for convergence
      If ITS = 30 Then  '!!!
        MsgBox "No convergence in 30 iterations for the SVDcmp routine"
        Stop
      End If
      '... shift from bottom 2-by-2 minor
      x = W(l)
      NM = k - 1
      y = W(NM)
      G = RV1(NM)
      H = RV1(k)
      f = ((y - Z) * (y + Z) + (G - H) * (G + H)) / (2! * H * y)
      G = Sqr(f * f + 1!)
      If f < 0 Then GsgnF = -Abs(G) Else GsgnF = Abs(G)
      f = ((x - Z) * (x + Z) + H * ((y / (f + GsgnF)) - H)) / x
      '... next QR transformation
      c = 1!
      S = 1!
      For j = l To NM
        i = j + 1
        G = RV1(i)
        y = W(i)
        H = S * G
        G = c * G
        Z = Sqr(f * f + H * H)
        RV1(j) = Z
        c = f / Z
        S = H / Z
        f = (x * c) + (G * S)
        G = -(x * S) + (G * c)
        H = y * S
        y = y * c
        For NM = 1 To n
          x = V(NM, j)
          Z = V(NM, i)
          V(NM, j) = (x * c) + (Z * S)
          V(NM, i) = -(x * S) + (Z * c)
        Next NM
        Z = Sqr(f * f + H * H)
        W(j) = Z
        If Z <> 0! Then
          '... note that rotation is arbitrary if Z=0
          Z = 1! / Z
          c = f * Z
          S = H * Z
        End If
        f = (c * G) + (S * y)
        x = -(S * G) + (c * y)
        For NM = 1 To M
          y = a(NM, j)
          Z = a(NM, i)
          a(NM, j) = (y * c) + (Z * S)
          a(NM, i) = -(y * S) + (Z * c)
        Next NM
      Next j
      RV1(l) = 0!
      RV1(k) = f
      W(k) = x
    Next ITS
  Next k

  Erase RV1
End Sub

Sub SvdInv(U() As Single, W() As Single, V() As Single, cNum As Single, M As Integer)
  '=============================================================================
  '... Calculates the inverse of matrix A() using the singular value
  ' decomposition matrices. The inverse is returned in U().
  ' Cnum is the condition number for matrix A().
  ' Dimensions of U(), W(), and V() are set in the main procedure.
  '=============================================================================
  Dim Ainv() As Single
  ReDim Ainv(1 To M, 1 To M)
  Dim WTI() As Single
  ReDim WTI(1 To M)
  Dim i As Integer, j As Integer, k As Integer

  '... Calculate the inverse of A()
  For i = 1 To M
    WTI(i) = 0!
    If W(i) <> 0 Then WTI(i) = 1 / W(i)
  Next i

  '... Ainv(n,n) = V(n,n) * [diag(1/W(n))] * [tranpose U(n,n)]
  ' Using indicial notation: Ainv(j,k) = V(j,i) * (1/W(i,i)) * U(k,i)
  For j = 1 To M
    For k = 1 To M
      Ainv(j, k) = 0
      For i = 1 To M
        Ainv(j, k) = Ainv(j, k) + V(j, i) * WTI(i) * U(k, i)
      Next i
    Next k
  Next j

  '... Load Ainv() into U()
  For i = 1 To M
    For j = 1 To M
      U(i, j) = Ainv(i, j)
    Next j
  Next i

  '... Calculate condition number
  Dim Wmax As Single
  Dim Wmin As Single

  Wmax = 0
  Wmin = 1E+25
  For j = 1 To M
    If W(j) > Wmax Then Wmax = W(j)
    If W(j) < Wmin Then Wmin = W(j)
  Next j
  If Wmin <> 0 Then cNum = Wmax / Wmin Else cNum = 0
  Erase WTI, Ainv
End Sub

Function TaufromBeta(Beta As Single) As Single
  '============================================================================
  '... This routine calculates the FT age from Beta = ln(Ns/Ni).
  ' TaufromBeta = FT age in Ma
  ' Zeta0 (shared) = zeta factor
  ' RhoD0 (shared) = monitor track density used for this dataset
  ' LamdaD (shared) = total decay constant for 238U (yr^-1)
  ' GF (shared) = geometry factor
  Dim BBeta As Single
  Dim ExpZm As Single

  Select Case Beta
    Case Is > 80
      BBeta = 80
    Case Is < -80
      BBeta = -80
    Case Else
      BBeta = Beta
  End Select
  ExpZm = LamdaD * Zeta0 * GF * RhoD0 * Exp(BBeta)
  TaufromBeta = log(1 + ExpZm) / (1000000! * LamdaD)
End Function

Public Function TaufromZ(Z As Single) As Single
  '============================================================================
  '... Converts to age in Myr from Z.
  ' LamdaD is a constant set in the main program.
  '============================================================================
  TaufromZ = log(1 + Exp(Z)) / (1000000# * LamdaD)
End Function

