VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "PD_plot"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

'class: PD_plot
'-------------

' --------------------- plot depended --------------------------------
'public
' methods:
'  show(PictureBox, CImode, peakMode, plotMode) - calculate and show PD plot
'  getPlotName()
'  getData() as string    - calculate and return data for PD plot (for save)

   'for MouseMove event
'  cPD(x,y) as string     - return current PD for stattus bar
'  cAge(x,y) as string    - return current Age for stattus bar
'
'private
' methods:
'  calc()                 - calculate data for PD plot, used by show() and getData()
'  PDmax() As Single      - used by show()
'  Sub Xscale(pic As PictureBox, Label As String, xmin As Single, xmax As Single)
'  Sub Lscale(pic As PictureBox, Label As String, ymin As Single, ymax As Single)
'  Sub Rscale(pic As PictureBox, Label As String, ymin As Single, ymax As Single)
'
' vars:
   Private cx As Single
   Private cy As Single
   Private xmin As Single
   Private ymin As Single
   Private xmax As Single
   Private ymax As Single
   Private pCount As Integer   'peaks
   Private bCount As Integer   'bars
   Private gCount As Integer   'dots (5+pCount)
   Private peak() As Single    'peak values
   Private bar() As Single     'bars values
   Private pd() As Single      'graphs values
   'labels:
   Private Const plotName = "Probability-Density Plot"
   Private Const xLabel = "FT grain age (Ma)"
   Private Const yLeft = "Density (% per delta z=0.1)"
   Private Const yRight = ""
   Private Const plotName2 = "Probability-Density Plot with Best-Fit Peaks"
   'colors:
   Private Const colPD = &HFF0000
   Private Const colObs = &HFF&
   Private Const colHist = &HC0C0FF
'
' --------------------- plot undepended --------------------------------
'public
' methods:
'  Sub init(pic As PictureBox)
'  Function getY0(pic As PictureBox)
'  Function getYY(pic As PictureBox)
'  Function getX0(pic As PictureBox)
'  Function getXX(pic As PictureBox)
'  Sub drawTitle(pic As PictureBox, name As String)
'  Sub drawAxis(pic As PictureBox)
'  Sub setScale(pic As PictureBox, x0 As Single, xx As Single, y0 As Single, yy As Single)
' vars:
   Private mode As Integer   'plot Mode
   'colors:
   Private Const colBack = &H80000005
   Private Const colFill = &H80000005
   'font:
   Private Const Font = "Times New Roman"
   Private Const screenFontSize = 10
   Private Const printFontSize = 14
   Private hFont As Single
   'window:
   Private psize As Single
   Private dh2 As Single
   Private dw2 As Single
   Private Const cScreen = 3
   Private Const cCopy = 2
   Private Const cPrint = 4
   Private Const cTopScreen = 0.17
   Private Const cTopCopy = 0.25
   Private Const cTopPrint = 0.38
'

' --------------------- plot depended methods --------------------------------
'
Public Function getPlotName(mode As Integer) As String
  If mode = 1 Then getPlotName = plotName Else getPlotName = plotName2
End Function

Public Function cPD(x As Single, y As Single) As String
  If x >= xmin And y >= ymin And x <= xmax And y <= ymax Then
    cx = Exp(x)
    cy = y
    cPD = "PD: " + Format(cy, "###0.0")
  Else
    cPD = ""
  End If
End Function

Public Function cAge(x As Single, y As Single) As String
  If x >= xmin And y >= ymin And x <= xmax And y <= ymax Then
    cx = Exp(x)
    cy = y
    cAge = "Age: " + Format(cx, "####0")
  Else
    cAge = ""
  End If
End Function

Private Sub calc()
  'Converted from ZetaAge
  '===========================================================================
  '... Calculates probability density and histogram data for plotting.
  ' Required shared variables are:
  ' Zgrain(), Zerr(), Num%:  Z, SE(Z), and number for grain ages
  ' Zerr() should be for grain-error only.
  ' QMeanZerr is the quadratic mean of Zerr().
  ' The total probability mass integrates to N (= number of grains).
  ' Histogram bars and probability density are given in
  ' number of grains for delta Z = 0.1.
  '===========================================================================
  Const BarWidth = 0.1   'width of histogram bar in Z units
  Const zWidth = BarWidth / 5 'interval width for PD plot in Z units
  Dim i As Integer, j As Integer
  Dim PkZ(1 To maxPeaks) As Single
  Dim Zmin As Single, Zmax As Single
  Dim LwLmt As Single, UpLmt As Single
  Dim Zi As Single, Zi2 As Single
  Dim Dsum As Single, SEDsum As Single
  Dim SumPkD As Single
  Dim dev As Single, S As Single
  Dim PkD As Single

  '... Set up peak list
  pCount = PkNum
  If PkNum > 0 Then
    ReDim peak(1 To PkNum, 1 To 3)
    For i = 1 To PkNum
      PkZ(i) = ZfromTau(PkAge(i))
      peak(i, 1) = PkAge(i)
      peak(i, 2) = PkSDz(i)
      peak(i, 3) = PkFrac(i) * Num
    Next i
  End If
  
  '... Find min, max, QMean of Zerr()
  Zmin = Zgrain(1)
  Zmax = Zgrain(1)
  For i = 2 To Num
    If Zgrain(i) > Zmax Then Zmax = Zgrain(i)
    If Zgrain(i) < Zmin Then Zmin = Zgrain(i)
  Next i

  '... Determine even limits for the histogram part of the plot.
  LwLmt = BarWidth * (Int(Zmin / BarWidth) - 1)
  UpLmt = BarWidth * (1 + Int(Zmax / BarWidth))
  bCount = Round((UpLmt - LwLmt) / BarWidth) + 1
  ReDim bar(1 To bCount, 1 To 2)
  Zi = LwLmt
'  For Zi = LwLmt To (UpLmt + 1) Step BarWidth
  For j = 1 To bCount
    '... Determine count for current histogram bar
    bar(j, 1) = TaufromZ(Zi)
    bar(j, 2) = 0
    Zi2 = Zi + BarWidth
    For i = 1 To Num
      If Zgrain(i) >= Zi And Zgrain(i) < Zi2 Then bar(j, 2) = bar(j, 2) + 1
    Next i
    bar(j, 2) = 100 * bar(j, 2) / Num
    Zi = Zi2
  Next j

  '... Calculate probability density distribution and write output data.
  '... Start and end at 5*QMeanZerr below Zmin and above Zmax, respectively.
  LwLmt = Zmin - (5 * QMeanZerr)
  UpLmt = Zmax + (5 * QMeanZerr)
  Zi = LwLmt
'  set above: zWidth = BarWidth / 5
  gCount = Round((UpLmt - LwLmt) / zWidth) + 1
  ReDim pd(1 To gCount, 1 To 5 + PkNum)
  For j = 1 To gCount
    '... Calculate probability density on the Z scale
    Call KernelPD(Zi, -1, Dsum, SEDsum)
    Dsum = 100 * BarWidth * Dsum / Num
    SEDsum = 100 * BarWidth * SEDsum / Num
    pd(j, 1) = TaufromZ(Zi)
    pd(j, 2) = Dsum
    pd(j, 3) = Dsum - SEDsum
    pd(j, 4) = Dsum + SEDsum

    SumPkD = 0
    For i = 1 To PkNum
      dev = Zi - PkZ(i)
      S = PkSDz(i)
      If Abs(dev) < 5 * S Then
        PkD = (100 * BarWidth * peak(i, 3) / Num) * (C0 / S) * Exp(-(dev * dev) / (2 * S * S))
      Else
        PkD = 0
      End If
      SumPkD = SumPkD + PkD
      pd(j, 4 + i) = PkD
    Next i
    pd(j, 4 + i) = SumPkD
    Zi = Zi + zWidth
  Next j
End Sub

Private Function PDmax() As Single
  Dim i As Integer, j As Integer, n As Integer
  Dim yy As Single

  yy = 0
  For i = 1 To bCount
    If yy < bar(i, 2) Then yy = bar(i, 2)
  Next i
  n = 5 + pCount
  For i = 1 To gCount
    For j = 2 To n
      If yy < pd(i, j) Then yy = pd(i, j)
    Next j
  Next i
  PDmax = yy
End Function

Public Function getData() As String
  Dim i As Integer, j As Integer
  Dim ss As String, H As String

  '... set up header string
  H = "LwEndofBar(Ma), Freq(%/deltaZ=0.1), Age(Ma), ObsPD(%/deltaZ=0.1), PD-SE(%/deltaZ=0.1), PD+SE(%/deltaZ=0.1)"
  calc  '... calculate grain data
  ss = "Kernel factor =  .6  (Ratio of kernel window size to standard error)" + vbCrLf
  For i = 1 To pCount
    ss = ss + "Peak #" + Str(i) + ". Mean Age, W in z units, Number of grains: " + nFormat(peak(i, 1), 1, 7) + "," + nFormat(peak(i, 2), 2, 6) + "," + nFormat(peak(i, 3), 1, 6) + vbCrLf
    H = H + ", PDpk" + LTrim$(Str$(i)) + "(%/deltaZ=0.1)"
  Next i

  ss = ss + H + ", PDpeaks(%/deltaZ=0.1)" + vbCrLf
  For j = 1 To gCount
    If j <= bCount Then
      ss = ss + nFormat(bar(j, 1), 2, 8) + "," + nFormat(bar(j, 2), 2, 8)
    Else
      ss = ss + " , "
    End If
    ss = ss + "," + nFormat(pd(j, 1), 2, 8) + "," + nFormat(pd(j, 2), 2, 8) + "," + nFormat(pd(j, 3), 2, 8) + "," + nFormat(pd(j, 4), 2, 8)
    For i = 1 To PkNum
      ss = ss + "," + nFormat(pd(j, 4 + i), 2, 8)
    Next i
    ss = ss + "," + nFormat(pd(j, 4 + i), 2, 8) + vbCrLf
  Next j
  getData = ss
End Function

' log scale
Private Sub Xscale(pic As PictureBox, Label As String, ByVal xmin As Single, ByVal xmax As Single)
  Dim dd As Single, xlog As Single, cx As Single
  Dim ss As String

  dd = 0.25 * hFont
  cx = CLng(Exp(xmin))
  While cx <= Exp(xmax)
    xlog = log(cx)
    pic.Line (xlog, ymin)-(xlog, ymax - dd)
    ss = Format(cx / 1000)
    pic.CurrentX = xlog - 0.6 * pic.TextWidth(ss)
    pic.CurrentY = ymin - dd
    pic.Print ss
    cx = cx * 2
  Wend
End Sub

Private Sub Lscale(pic As PictureBox, Label As String, ByVal ymin As Single, ByVal ymax As Single)
  Dim y As Single, dy As Single, ds As Single, dd As Single, dh As Single
  Dim cx As Single, cy As Single, i As Single
  Dim yText As New Rotator

  y = ymax - ymin
  dy = 1
  While y >= 10 * dy
    dy = dy * 2
  Wend
  ds = psize / ymax * dy
  dd = 0.5 * Abs(pic.TextWidth("0"))
  dh = 0.5 * hFont
  cx = getX0(pic)
  cy = getY0(pic)
  For i = ymin To ymax Step dy
    pic.Line (cx - dd, cy)-(cx, cy)
    pic.CurrentX = cx - dd - pic.TextWidth(Str(CInt(i)) + " ")
    pic.CurrentY = cy + dh
    pic.Print CInt(i)
    cy = cy + ds
  Next i
End Sub

Private Sub Rscale(pic As PictureBox, Label As String, ByVal ymin As Single, ByVal ymax As Single)
  Dim y As Single, dy As Single, ds As Single, dd As Single, dh As Single
  Dim cx As Single, cy As Single, i As Single
  Dim yText As New Rotator

  y = ymax - ymin
  dy = 1
  While y >= 10 * dy
    dy = dy * 2
  Wend
  ds = psize / ymax * dy
  dd = 0.5 * Abs(pic.TextWidth("0"))
  dh = 0.5 * hFont
  cx = getXX(pic)
  cy = getY0(pic)
  For i = ymin To ymax Step dy
    pic.Line (cx, cy)-(cx + dd, cy)
    pic.CurrentX = cx + dd
    pic.CurrentY = cy + dh
    pic.Print CInt(i)
    cy = cy + ds
  Next i
End Sub

Public Sub show(pic As PictureBox, _
                CImode As Integer, _
                peakMode As Integer, _
                plotMode As Integer _
                )

  Dim i As Single, n As Integer, j As Integer
  Dim N1 As Integer, N2 As Integer
  Dim x0 As Single, y0 As Single, xx As Single, yy As Single
  Dim x As Single, y As Single, cx As Single

  mode = plotMode   ' 0-Screen, 1-Copy, 2-Print
  calc
  If bar(1, 1) <= 1 Then
    x0 = log(Int(bar(1, 1) * 10) * 100)
  ElseIf bar(1, 1) <= 10 Then
    x0 = log(Int(bar(1, 1)) * 1000)
  Else
    x0 = log(Int(bar(1, 1) / 10) * 10000)
  End If
  xx = log(CLng(bar(bCount, 1) * 1200))

  y0 = 0
  yy = PDmax() * 1.05
  x = xx - x0
  y = yy - y0

  init pic
  If mode = 2 Then
    If peakMode = 1 Then
      drawTitle pic, plotName
    Else
      drawTitle pic, plotName2
    End If
  End If
  drawAxis pic
  Lscale pic, yLeft, y0, yy
  Rscale pic, yRight, y0, yy
  setScale pic, x0, xx, y0, yy
  Xscale pic, xLabel, x0, xx

  'histogram
  For i = 1 To bCount - 1
    pic.Line (log(bar(i, 1) * 1000), 0)-(log(bar(i + 1, 1) * 1000), bar(i, 2)), colHist, BF
  Next i

  'PD
  i = 2
  n = 5 + pCount
  Do While log(pd(i, 1) * 1000) <= xx
    cx = log(pd(i - 1, 1) * 1000)
    If cx >= x0 Then
      If CImode = 0 Then
        pic.Line (cx, pd(i - 1, 3))-(log(pd(i, 1) * 1000), pd(i, 3)), colObs
        pic.Line (cx, pd(i - 1, 4))-(log(pd(i, 1) * 1000), pd(i, 4)), colObs
      Else
        pic.Line (cx, pd(i - 1, 2))-(log(pd(i, 1) * 1000), pd(i, 2)), colObs
      End If
      If PkNum > 0 Then
        If peakMode = 1 Then
          pic.Line (cx, pd(i - 1, n))-(log(pd(i, 1) * 1000), pd(i, n)), colPD
        Else
          For j = 5 To n - 1
            pic.Line (cx, pd(i - 1, j))-(log(pd(i, 1) * 1000), pd(i, j)), colPD
          Next j
        End If
      End If
    End If
    i = i + 1
    If i > gCount Then Exit Do
  Loop
End Sub

'------------------------ plot undepended methods --------------------------------

Private Sub init(pic As PictureBox)
  Dim hh As Single

  pic.Cls
  pic.Font = Font
  If mode = 0 Then
    pic.Font.Size = screenFontSize
  Else
    pic.Font.Size = printFontSize
  End If
  pic.FillStyle = 0
  pic.BackColor = colBack
  pic.FillColor = colFill
  pic.ScaleMode = 0    'custom
  pic.Scale (0, pic.Height)-(pic.Width, 0)
  hh = Abs(pic.ScaleHeight)
  hFont = Abs(pic.TextHeight("0"))
  If mode = 0 Then                                  'screen
    psize = hh - cScreen * hFont
  ElseIf mode = 1 Then                              'copy
    psize = hh - cCopy * hFont
  ElseIf mode = 2 Then                              'print
    psize = hh - cPrint * hFont
  End If
  dh2 = (hh - psize) / hh
  dw2 = (pic.ScaleWidth - psize) / pic.ScaleWidth
End Sub
  
Private Function getY0(pic As PictureBox)
  Dim hh As Single

  hh = Abs(pic.ScaleHeight)
  If mode = 0 Then
    getY0 = pic.ScaleTop - hh + (hh * dh2) * (1 - cTopScreen)
  ElseIf mode = 1 Then
    getY0 = pic.ScaleTop - hh + (hh * dh2) * (1 - cTopCopy)
  ElseIf mode = 2 Then
    getY0 = pic.ScaleTop - hh + (hh * dh2) * (1 - cTopPrint)
  End If
End Function

Private Function getYY(pic As PictureBox)
  Dim hh As Single

  hh = Abs(pic.ScaleHeight)
  If mode = 0 Then
    getYY = pic.ScaleTop - (hh * dh2) * cTopScreen
  ElseIf mode = 1 Then
    getYY = pic.ScaleTop - (hh * dh2) * cTopCopy
  ElseIf mode = 2 Then
    getYY = pic.ScaleTop - (hh * dh2) * cTopPrint
  End If
End Function

Private Function getX0(pic As PictureBox)
  getX0 = pic.ScaleLeft + (pic.ScaleWidth * dw2) / 2
End Function

Private Function getXX(pic As PictureBox)
  getXX = pic.ScaleLeft + pic.ScaleWidth - (pic.ScaleWidth * dw2) / 2
End Function

Private Sub drawTitle(pic As PictureBox, name As String)
  pic.Font.Size = 1.1 * printFontSize
  pic.Font.Bold = True
  pic.CurrentX = (pic.ScaleWidth - pic.TextWidth(name)) / 2
  pic.CurrentY = pic.ScaleTop
  pic.Print name
  pic.Font.Size = printFontSize
  pic.Font.Bold = False
End Sub

Private Sub drawAxis(pic As PictureBox)
  Dim yText As New Rotator

  pic.Line (getX0(pic), getY0(pic))-(getXX(pic), getYY(pic)), , B
  If mode <> 1 Then
    'X
    pic.CurrentX = pic.ScaleLeft + (pic.ScaleWidth - pic.TextWidth(xLabel)) / 2
    pic.CurrentY = 1.2 * hFont
    pic.Print xLabel
    ' Left
    Set yText.Device = pic
    yText.Angle = 90
    pic.CurrentY = getY0(pic) + (psize - pic.TextWidth(yLeft)) / 2
    pic.CurrentX = 0.2 * hFont
    yText.Label yLeft
    ' Right
    Set yText.Device = pic
    yText.Angle = 90
    pic.CurrentY = getY0(pic) + (psize - pic.TextWidth(yRight)) / 2
    pic.CurrentX = pic.ScaleWidth - 1.2 * hFont
    yText.Label yRight
  End If
End Sub

Private Sub setScale(pic As PictureBox, ByVal x0 As Single, ByVal xx As Single, ByVal y0 As Single, ByVal yy As Single)
  Dim dx As Single, dy As Single

  dx = (xx - x0) * dw2 / (1 - dw2)
  dy = (yy - y0) * dh2 / (1 - dh2)
  If mode = 0 Then
    pic.Scale (x0 - dx / 2, yy + cTopScreen * dy)-(xx + dx / 2, y0 - (1 - cTopScreen) * dy)
  ElseIf mode = 1 Then
    pic.Scale (x0 - dx / 2, yy + cTopCopy * dy)-(xx + dx / 2, y0 - (1 - cTopCopy) * dy)
  ElseIf mode = 2 Then
    pic.Scale (x0 - dx / 2, yy + cTopPrint * dy)-(xx + dx / 2, y0 - (1 - cTopPrint) * dy)
  End If
  hFont = Abs(pic.TextHeight("0"))
End Sub

