Attribute VB_Name = "ZetaAge"
'Converted from ZetaAge

Option Explicit

Public Const PFmax = 100       'max potential peaks
Public Ppeaks() As Double      ' potential peaks
Public Pnum As Integer         ' number of potential peaks



Public Sub getPpeaks()
  Dim i As Integer
  If Num > 0 Then
    toLogLine "# Potential peaks:"
'    Pnum = FindPeaks(1, 1)    ' first derivative list, 63%
    Pnum = FindPeaks(2, 1)   ' second derivative list, 63%
    Pnum = peaksCorrection()
    toLogLine "         Age       PD    Est.N"
    For i = 1 To Pnum
      toLogLine nFormat(i, 0, 3) + nFormat(Ppeaks(i, 1), 2, 9) + nFormat(Ppeaks(i, 2), 2, 9) + nFormat(Ppeaks(i, 3), 2, 9)
    Next i
    toLogLine "  ----------------------------"
    toLogLine "  Pool Age: " + nFormat(Ppeaks(0, 1), 2, 9) + " Ma"
    toLogLine ""
  End If
End Sub



Private Function peaksCorrection() As Integer
  Dim i As Integer
  Dim n As Integer
  Dim k As Integer
  Dim done As Integer

  n = 0
  k = 0
  done = 0
  For i = 1 To Pnum
    If CInt(Ppeaks(i, 1)) <= CInt(AgeMin) Then
      If k = 0 Then
        k = 1
        n = n + 1
      End If
      If Ppeaks(i, 1) < AgeMin Then
        Ppeaks(n, 1) = AgeMin
      Else
        Ppeaks(n, 1) = Ppeaks(i, 1)
      End If
    ElseIf CInt(Ppeaks(i, 1)) < CInt(AgeMax) Then
      n = n + 1
      Ppeaks(n, 1) = Ppeaks(i, 1)
    Else
      n = n + 1
      If Ppeaks(i, 1) < AgeMax Then Ppeaks(n, 1) = Ppeaks(i, 1) Else Ppeaks(n, 1) = AgeMax
      done = 1
    End If
    Ppeaks(n, 2) = Ppeaks(i, 2)
    Ppeaks(n, 3) = Ppeaks(i, 3)
    If done = 1 Then Exit For
  Next i
  peaksCorrection = n
End Function


Private Function FindPeaks(dmode As Integer, mode As Integer) As Integer
  Dim Pnum As Integer
  Dim amin, amax
  Dim SumNs As Double
  Dim SumNi As Double
  Dim i As Integer
  Dim PoolAge As Double, PoolAgeL As Double, PoolAgeU As Double

  ReDim Ppeaks(0 To PFmax, 0 To 3)

  If dmode = 1 Then
    Pnum = PkFind1(amin, amax)
  ElseIf dmode = 2 Then
    Pnum = PkFind2(amin, amax)
  Else
    Pnum = PkFind2(amin, amax)
    If Pnum > PFmax Then Pnum = PkFind1(amin, amax)
    If Pnum > PFmax Then Pnum = PFmax
  End If

  '... Calculate age and uncertainties for pooled age
  SumNs = 0
  SumNi = 0
  For i = 1 To Num
    SumNs = SumNs + Ns(i)
    SumNi = SumNi + Ni(i)
  Next i
  Call FTeqn(SumNs, SumNi, RhoD0, RERhoD0, Zeta0, REZeta0, mode, PoolAge, PoolAgeL, PoolAgeU)
  Ppeaks(0, 1) = PoolAge
  FindPeaks = Pnum
End Function



Private Function PkFind1(AgeMin, AgeMax) As Integer
  'output: AgeMin, AgeMax
  ' return number of peaks
  '========================================================================
  '... Finds peaks in the probability density distribution calculated from
  ' information stored in the following shared variables:
  ' FTage(), Num%:     fission track grain ages and number of grain ages
  ' Zgrain(), Zerr():  Z and SE(Z) for the grains
  ' Zerr() should be in units of 1*SE and should not include uncertainties
  ' for monitor track density and zeta factor.

  Dim ZUp(1 To maxGrains) As Double
  Const Zacc = 0.003    '... specified accuracy for Z at peak
  Dim PkNum As Integer
  Dim Zmin As Double, Zmax As Double
  Dim dZ0 As Double, dz As Double
  Dim Flw As Double
  Dim Dum As Double
  Dim Fup As Double
  Dim Zi As Double
  Dim j As Integer, i As Integer
  '========================================================================
  '... Find min and max
  Zmin = Zgrain(1)
  Zmax = Zgrain(1)
  For j = 2 To Num
    If Zgrain(j) > Zmax Then Zmax = Zgrain(j)
    If Zgrain(j) < Zmin Then Zmin = Zgrain(j)
  Next j
  AgeMax = TaufromZ(Zmax)
  AgeMin = TaufromZ(Zmin)

  '... To catch the tails of the distribution, the range for Z is set
  ' at 2 standard errors below and above the minimum and maximum Zgrain.
  Zmax = Zmax + 2 * QMeanZerr
  Zmin = Zmin - 2 * QMeanZerr
  '... C0 is defined as a constant in the main program.
  '... The uncertainty for each grain age is the grain-only standard error.
  PkNum = 0
  dZ0 = (Zmax - Zmin) / Num
  Call KernelPD(Zmin, 1, Flw, Dum)
  For Zi = (Zmin + dZ0) To Zmax Step dZ0
    Call KernelPD(Zi, 1, Fup, Dum)
    '... Maximum in the density function is indicated by the 1rd derivative
    ' changing from positive to negative with increasing Zi.
    If Flw > 0 And Fup <= 0 Then
      PkNum = PkNum + 1
'      ZLw(PkNum) = Zi - dZ
      ZUp(PkNum) = Zi
    End If
    Flw = Fup
    If PkNum = Num Then Exit For
  Next Zi

  Dim UpZ As Double
  Dim LwZ As Double
  Dim f As Double
  Dim Fmid As Double
  Dim RTbis As Double
  Dim MidZ As Double
  Dim PkDen As Double
  Dim PkArea As Double
  Dim pkAge As Double
  Dim n As Integer
  n = 0
  For i = 1 To PkNum
    UpZ = ZUp(i)
    LwZ = ZUp(i) - dZ0
    Call KernelPD(LwZ, 1, f, Dum)
    Call KernelPD(UpZ, 1, Fmid, Dum)
    If f <= 0 Or Fmid >= 0 Then                   '!!!
       MsgBox "Incorrect brackets for bisection"
       Stop
    End If
    RTbis = UpZ
    dz = LwZ - UpZ
    For j = 1 To 40
      dz = dz * 0.5
      MidZ = RTbis + dz
      Call KernelPD(MidZ, 1, Fmid, Dum)
      If Fmid <= 0 Then RTbis = MidZ
      If (Abs(dz) < Zacc) Or (Fmid = 0) Then Exit For
    Next j
    If j = 40 Then MsgBox "Too many bisections"     '!!!

    '... Calculate peak density and peak area
    ' Peak area = peak density * std dev *Sqr(2*Pi)
    Call KernelPD(RTbis, 0, PkDen, Dum)
    PkArea = PkDen * QMeanZerr / C0
    PkDen = 0.1 * PkDen

    '... Convert from the Z scale to the Tau scale.
    pkAge = TaufromZ(RTbis)

    '... Record location of peak
    If PkNum <= PFmax Then
      n = n + 1
      Ppeaks(n, 1) = pkAge
      Ppeaks(n, 2) = PkDen
      Ppeaks(n, 3) = PkArea
    End If
  Next i
  PkFind1 = n
End Function



Private Function PkFind2(AgeMin, AgeMax) As Integer
  'output: AgeMin, AgeMax, PkNum(global)
  'return number of peaks
  '========================================================================
  '... Finds peaks in the probability density distribution calculated from
  ' information stored in the following shared variables:
  ' FTage(), Num%:     fission track grain ages and number of grain ages
  ' Zgrain(), Zerr():  Z and SE(Z) for the grains
  ' Zerr() should be in units of 1*SE and should not include uncertainties
  ' for monitor track density and zeta factor.
  '========================================================================
  Dim ZUp(1 To maxGrains) As Double
  Const Zacc = 0.003    '... specified accuracy for Z at peak
  Dim PkNum As Integer
  Dim Zmin As Double, Zmax As Double
  Dim dZ0 As Double, dz As Double
  Dim Flw As Double
  Dim Dum As Double
  Dim Fup As Double
  Dim Zi As Double
  Dim j As Integer, i As Integer

  '... Find min and max
  Zmin = Zgrain(1)
  Zmax = Zgrain(1)
  For j = 2 To Num
    If Zgrain(j) > Zmax Then Zmax = Zgrain(j)
    If Zgrain(j) < Zmin Then Zmin = Zgrain(j)
  Next j
  AgeMax = TaufromZ(Zmax)
  AgeMin = TaufromZ(Zmin)

  '... To catch the tails of the distribution, the range for Z is set
  ' at 2 standard errors below and above the minimum and maximum Zgrain.
  Zmax = Zmax + 2 * QMeanZerr
  Zmin = Zmin - 2 * QMeanZerr

  PkNum = 0
  dZ0 = (Zmax - Zmin) / Num
  Call KernelPD(Zmin, 3, Flw, Dum)
  For Zi = (Zmin + dZ0) To Zmax Step dZ0
    Call KernelPD(Zi, 3, Fup, Dum)
    '... Minimum in the 2nd derivative is indicated by the 3rd derivative
    ' changing from negative to positive with increasing Zi.
    If Flw < 0 And Fup > 0 Then
      PkNum% = PkNum% + 1
      ZUp(PkNum%) = Zi
    End If
    Flw = Fup
    If PkNum% = Num% Then Exit For
  Next Zi

  Dim UpZ As Double
  Dim LwZ As Double
  Dim f As Double
  Dim Fmid As Double
  Dim RTbis As Double
  Dim MidZ As Double
  Dim PkDen As Double
  Dim PkArea As Double
  Dim pkAge As Double
  Dim n As Integer
  n = 0
  For i% = 1 To PkNum%
    UpZ = ZUp(i%)
    LwZ = UpZ - dZ0
    Call KernelPD(LwZ, 3, f, Dum)
    Call KernelPD(UpZ, 3, Fmid, Dum)
    If f >= 0 Or Fmid <= 0 Then               '!!!
      MsgBox "Incorrect brackets for bisection"
      Stop
    End If
    RTbis = LwZ
    dz = UpZ - LwZ
    For j = 1 To 40
      dz = dz * 0.5
      MidZ = RTbis + dz
      Call KernelPD(MidZ, 3, Fmid, Dum)
      If Fmid <= 0 Then RTbis = MidZ
      If (Abs(dz) < Zacc) Or (Fmid = 0) Then Exit For
    Next j
    If j% = 40 Then MsgBox "Too many bisections"     '!!!

    '... Calculate peak density and peak area
    ' Peak area = peak density * std dev *Sqr(2*Pi)
    Call KernelPD(RTbis, 0, PkDen, Dum)
    PkArea = PkDen * QMeanZerr / C0
    PkDen = 0.1 * PkDen

    '... Convert from the Z scale to the Tau scale.
    pkAge = TaufromZ(RTbis)

    '... Record location of peak
    If PkNum <= PFmax Then
      n = n + 1
      Ppeaks(n, 1) = pkAge
      Ppeaks(n, 2) = PkDen
      Ppeaks(n, 3) = PkArea
    End If
  Next i
  PkFind2 = n
End Function



Private Sub FTeqn(Ns As Double, Ni As Double, RhoD As Double, RERhoD As Double, Zeta As Double, REZeta As Double, CIOpt As Integer, age As Double, AgeL As Double, AgeU As Double)
  '============================================================================
  '... Calculates FT age and confidence interval and includes uncertainties
  ' for spontaneous, induced, and monitor track densities, and the zeta factor.
  ' The FT age is calculated using the Zeta method equation of Hurford and
  ' Green (Isotope Geoscience, 1983) and the formulation of Galbraith
  ' (Nuclear Tracks, 1990). Confidence intervals are determined using
  ' the method outlined in Bardsley (Mathematical  Geology, 1983) and Galbraith
  ' (Mathematical Geology, 1984; ibid, 1990). When Ns or Ni is small, a median
  ' estimate and 95% confidence interval are calculated for the parameter
  ' (Ns*Nd/Ni) using Sneyd's algorithm (Computers and Geosciences, 1984).
  ' Constants LamdaD, and GF are set in main program. The geometric
  ' correction for the fluence monitor is explictly included as GF, so that the
  ' induced tracks (Ni) should be given as the actual number of counted tracks.
  '
  'INPUT:  Ns, Ni = spontaneous and induced tracks.
  '        RhoD, RErhodD = monitor track density and relative standard error.
  '        Zeta, REzeta = zeta factor and relative standard error.
  '        CIOpt% = option for confidence interval: 1 = 63%, 2 = 95%
  'OUTPUT: Mean age and 95% confidence interval in Ma.
  '============================================================================
  Dim Zfactor As Double
  Dim Alpha As Double
  Dim SE As Double
  Dim ExpZm As Double, ExpZ1 As Double, ExpZ2 As Double
  Dim ND As Double, N1 As Integer, N2 As Integer, N3 As Integer
  Dim CtArea As Double
  Dim ExpZAm As Double, ExpZA1 As Double, ExpZA2 As Double

  '... Select confidence interval
  Select Case CIOpt
    Case 1  '... 63% confidence interval
      Zfactor = 1
      Alpha = (1 - 0.63) / 2
    Case 2  '... 95% confidence interval
      Zfactor = 1.96
      Alpha = (1 - 0.95) / 2
  End Select

  '... Test for applicability of normal-distribution approximation.
  ' My experience is that the normal approximation can be used with the
  ' Z transform when SE(Z) < 0.15.
  SE = Sqr(1 / (Ns + 0.5) + 1 / (Ni + 0.5) + RERhoD * RERhoD + REZeta * REZeta)
  If SE < 0.15 Then
    '... Use the normal distribution with the Z transform to find
    ' confidence limits.
    ExpZm = LamdaD * Zeta * GF * RhoD * (Ns + 0.05) / (Ni + 0.5)
    ExpZ1 = ExpZm * Exp(-Zfactor * SE)
    ExpZ2 = ExpZm * Exp(Zfactor * SE)
  Else
    '... Find exact confidence limits for cases where Ns or Ni is small.
    ' Zeta uncertainties are assumed to be relatively small and therefore
    ' are not included in this calculation.
    '... Estimate equivalent track count for effective monitor track density
    If RERhoD <> 0 Then
        ND = 1 / (RERhoD * RERhoD)
    Else
        ND = 10000
    End If
    CtArea = ND / RhoD
    ND = CInt(ND)
    N1 = CInt(Min(Ns, ND))
    N2 = CInt(Max(Ns, ND) - N1%)
    N3 = CInt(N1% + Ni)
    '... Calculate mean and initial estimates of confidence limits
    SE = (Ns / ((Ns + 0.5) * (Ns + 0.5))) + (Ni / ((Ni + 0.5) * (Ni + 0.5)))
    SE = Sqr(SE + RERhoD * RERhoD)
    ExpZAm = (ND + 0.5) * (Ns + 0.5) / (Ni + 0.5)
    ExpZA1 = ExpZAm * Exp(-Zfactor * SE)
    ExpZA2 = ExpZAm * Exp(Zfactor * SE)
    '... Solve for exact confidence limits
    Call Solve(Alpha, N1, N2, N3, ExpZA2, ExpZ2, 1)
    Call Solve(0.5, N1, N2, N3, ExpZAm, ExpZm, 2)
    N1 = Max(0!, (N1 - 1!))
    Call Solve(1! - Alpha, N1, N2, N3, ExpZA1, ExpZ1, 1)
    ExpZ1 = LamdaD * Zeta * GF * ExpZ1 / CtArea
    ExpZm = LamdaD * Zeta * GF * ExpZm / CtArea
    ExpZ2 = LamdaD * Zeta * GF * ExpZ2 / CtArea
  End If
  AgeL = log(1 + ExpZ1) / (LamdaD * 1000000!)
  age = log(1 + ExpZm) / (LamdaD * 1000000!)
  AgeU = log(1 + ExpZ2) / (LamdaD * 1000000!)
End Sub



Private Sub Solve(a As Double, N1 As Integer, N2 As Integer, N3 As Integer, ThetaA As Double, Theta As Double, mode As Integer)
  '============================================================================
  '... This routine solves: FN(Theta) = Log(A)
  ' where ThetaA is the initial guess.
  '============================================================================
  Dim Al As Double
  Dim SStep As Double
  Dim Theta1 As Double, Theta2 As Double
  Dim S2 As Double
  Dim Ptest As Double

  Al = log(1 / a)
  '... Find an interval containing the solution
  Theta = ThetaA
  SStep = Sign((Theta / 10), (Xzdy(Theta, N1, N2, N3, mode) + Al))

  Theta1 = Theta + SStep
  Theta2 = Theta
  S2 = Sign(1, SStep)
  Ptest = Xzdy(Theta1, N1, N2, N3, mode) + Al
  Do While S2 = Sign(1, Ptest)
    Theta1 = Theta1 + SStep
    If Theta1 <= 0 Then
      Theta1 = 0.5 * (Theta1 - SStep)
      SStep = -0.2 * (Theta1 - SStep)
    End If
    Ptest = Xzdy(Theta1, N1, N2, N3, mode) + Al
  Loop

  '... Carry out bisection
  Do While Abs(Theta1 - Theta2) > (Max(1!, Theta1) * Errn)
    Theta = 0.5 * (Theta1 + Theta2)
    If Sign(1, Xzdy(Theta, N1, N2, N3, mode) + Al) = S2 Then
      Theta2 = Theta
    Else
      Theta1 = Theta
    End If
  Loop
End Sub





Private Function Max(a As Double, b As Double) As Double
  If a > b Then Max = a Else Max = b
End Function



Private Function Min(a As Double, b As Double) As Double
  If a < b Then Min = a Else Min = b
End Function


