VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "RadialPlot"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

'class RadialPlot
'----------------

' --------------------- plot depended --------------------------------
'public
' methods:
'  show(PictureBox, plotMode)       - calculate and show radial plot
'  getPlotName()
'  getData() as string    - calculate and return data for radial plot (for save)
'
'private
' methods:
'  calc()            - calculate data for radial plot, used by show() and getData()
'  Sub Xscale(pic As PictureBox, Label As String, xmin As Single, xmax As Single)
'  Sub Lscale(pic As PictureBox, Label As String, ymin As Single, ymax As Single)
'  Sub Rscale(pic As PictureBox, Label As String, ymin As Single, ymax As Single)
'
'  vars:
   Private Zmean As Single
   Private xmin As Single
   Private xmax As Single
   Private ymin As Single
   Private ymax As Single
   Private data() As Single
   Private Label(1 To 11, 1 To 5) As Single
   Private YmaxforPks() As Single
   'labels:
   Private Const plotName = "Radial Plot"
   Private Const xLabel = "SE[z] (~RSE[age] )"
   Private Const yLeft = "Standardized variance"
   Private Const yRight = "FT grain age (Ma)"
   'colors:
   Private Const colGrain = &HFF0000
   Private Const colPeak = &HFF&
'
' --------------------- plot undepended --------------------------------
'private
' methods:
'  Sub init(pic As PictureBox)
'  Function getY0(pic As PictureBox)
'  Function getYY(pic As PictureBox)
'  Function getX0(pic As PictureBox)
'  Function getXX(pic As PictureBox)
'  Sub drawTitle(pic As PictureBox, name As String)
'  Sub drawAxis(pic As PictureBox)
'  Sub setScale(pic As PictureBox, x0 As Single, xx As Single, y0 As Single, yy As Single)
'
'  vars:
   Private mode As Integer   'plot Mode
   'colors:
   Private Const colBack = &H80000005
   Private Const colFill = &H80000005
   'font:
   Private Const Font = "Times New Roman"
   Private Const screenFontSize = 10
   Private Const printFontSize = 14
   Private hFont As Single
   'window:
   Private psize As Single
   Private dh2 As Single
   Private dw2 As Single
   Private Const cScreen = 3
   Private Const cCopy = 2
   Private Const cPrint = 4
   Private Const cTopScreen = 0.17
   Private Const cTopCopy = 0.25
   Private Const cTopPrint = 0.38

' --------------------- plot depended methods --------------------------------
'
Public Function getPlotName() As String
  getPlotName = plotName
End Function

Private Sub calc()
  'Converted from ZetaAge
  '===========================================================================
  '... Calculates data for radial plot fo Galbraith (1990).
  ' Required shared variables are:
  ' Zgrain(), Zerr(), Num:  Z, SE(Z), and number for grain ages
  ' Zerr() should be for grain-error only.

  Dim j As Integer
  Dim SumZInvVar As Single
  Dim ZInvVar As Single
  Dim yRight As Single
  Dim Xstep As Single
  Dim Ystep As Single

  ReDim data(1 To Num, 1 To 2)
  '... Find min, max, weighted mean of Zgrain()
  SumZInvVar = 0
  Zmean = 0
  For j = 1 To Num
    ZInvVar = 1 / (Zerr(j) * Zerr(j))
    SumZInvVar = SumZInvVar + ZInvVar
    Zmean = Zmean + Zgrain(j) * ZInvVar
  Next j
  Zmean = Zmean / SumZInvVar
  xmin = 0
  xmax = 0
  For j = 1 To Num
    data(j, 1) = 1 / Zerr(j)
    data(j, 2) = (Zgrain(j) - Zmean) * data(j, 1)
    If data(j, 1) > xmax Then xmax = data(j, 1)
  Next j
  xmax = 1.05 * xmax
  Xstep = xmax / 10
  ymin = 10000000000#
  ymax = -10000000000#
  For j = 1 To Num
    yRight = (Zgrain(j) - Zmean) * xmax
    If yRight > ymax Then ymax = yRight
    If yRight < ymin Then ymin = yRight
  Next j
  ymin = CInt(ymin - 1)
  ymax = CInt(ymax + 1)
  If Abs(ymax) > Abs(ymin) Then ymin = -ymax Else ymax = -ymin
  Ystep = ymax / 5
  
  For j = 1 To 11
      Label(j, 1) = Ystep * (j - 6)                                     'Ytic
      Label(j, 2) = Zmean + Label(j, 1) / xmax                       'Ztic
      Label(j, 3) = log(1 + Exp(Label(j, 2))) / (1000000! * LamdaD)  'TauTic
      If j > 1 Then
        Label(j, 4) = (j - 1) * Xstep        'Xtic
        Label(j, 5) = 1 / Label(j, 4)     'SEZtic
      End If
  Next j
  
  '... lines for peaks
  ReDim YmaxforPks(PkNum)
  For j = 1 To PkNum
    YmaxforPks(j) = (ZfromTau(PkAge(j)) - Zmean) * xmax
  Next j

End Sub

Public Function getData() As String
  Dim j As Integer
  Dim ss As String
  
  calc  '... calculate grain ages
  ss = "X-Y DIMENSIONS FOR RADIAL PLOT" + vbCrLf
  ss = ss + "Zmean, Xmin, Xmax, Ymin, Ymax" + vbCrLf
  ss = ss + nFormat(Zmean, 3, 8) + ",  " + nFormat(xmin, 3, 8) + ",  " + nFormat(xmax, 3, 8) + ",  " + nFormat(ymin, 3, 8) + ",  " + nFormat(ymax, 3, 8) + vbCrLf
  ss = ss + vbCrLf + "GRAIN DATA, TIC POSITIONS, AND TIC LABELS" + vbCrLf
  ss = ss + "Grain_X_var, Grain_Y_Var, YpositionforTicLabels, ZTicLabelsforYaxis, AgeTicLabelsforYaxis, XpositionforTicLabels, SEZTicLabelsforXaxis" + vbCrLf

  For j = 1 To Num
    ss = ss + nFormat(data(j, 1), 3, 8) + ",  " + nFormat(data(j, 2), 3, 8)
    Select Case j
      Case 1
        ss = ss + ",  " + nFormat(Label(j, 1), 3, 8) + ",  " + nFormat(Label(j, 2), 3, 8) + ",  " + nFormat(Label(j, 3), 1, 6)
        ss = ss + ", " + nFormat(0, 3, 8) + ",   --" + vbCrLf
      Case 2 To 11
        ss = ss + ",  " + nFormat(Label(j, 1), 3, 8) + ",  " + nFormat(Label(j, 2), 3, 8) + ",  " + nFormat(Label(j, 3), 1, 6)
        ss = ss + ", " + nFormat(Label(j, 4), 3, 8) + ",  " + nFormat(Label(j, 5), 3, 6) + vbCrLf
      Case Else
        ss = ss + vbCrLf
    End Select
  Next j
  If PkNum > 0 Then
    ss = ss + vbCrLf + "LINE SEGMENTS FOR BEST-FIT PEAKS" + vbCrLf
    ss = ss + "Xval_Pks, YvalPks" + vbCrLf
    For j = 1 To PkNum
      ss = ss + nFormat(xmin, 3, 8) + ", " + nFormat(0, 3, 8) + vbCrLf
      ss = ss + nFormat(xmax, 3, 8) + ", " + nFormat(YmaxforPks(j), 3, 8) + vbCrLf
      ss = ss + vbCrLf
    Next j
  End If
  getData = ss
End Function

Private Sub Xscale(pic As PictureBox)
  Dim i As Integer, dd As Single, cy As Single, ss As String

  dd = 0.25 * Abs(pic.TextHeight("0"))
  cy = getY0(pic)
  For i = 1 To 10 Step 3
    pic.Line (i, cy)-(i, cy - dd)
    ss = nFormat(Label(i + 1, 5), 2, 4)
    pic.CurrentX = i - 0.6 * pic.TextWidth(ss)
    pic.CurrentY = cy - dd
    pic.Print ss
  Next i
End Sub

Private Sub Lscale(pic As PictureBox, ByVal ymin As Single, ByVal ymax As Single)
  Dim i As Integer, dy As Integer
  Dim dd As Single, dh As Single, cx As Single
  Dim yText As New Rotator

  dy = CInt((ymax - ymin) / 10)
  If dy = 0 Then dy = 1
  dd = 0.5 * Abs(pic.TextWidth("0"))
  dh = 0.5 * pic.TextHeight("0")
  cx = getX0(pic)
  i = 0
  While i <= ymax
    pic.Line (cx - dd, i)-(cx, i)
    pic.CurrentX = cx - dd - pic.TextWidth(Str(i) + " ")
    pic.CurrentY = i - dh
    pic.Print Str(i)
    pic.Line (cx - dd, -i)-(cx, -i)
    pic.CurrentX = cx - dd - pic.TextWidth(Str(-i) + " ")
    pic.CurrentY = -i - dh
    pic.Print Str(-i)
    i = i + dy
  Wend
End Sub

Private Sub Rscale(pic As PictureBox, ByVal ymin As Single, ByVal ymax As Single)
  Dim i As Integer, dh As Single, dd As Single, xx As Single

  xx = getXX(pic)
  dd = 0.5 * Abs(pic.TextWidth("0"))
  dh = 0.5 * pic.TextHeight("0")
  For i = 1 To 11
    pic.Line (xx, log(Label(i, 3)))-(xx + dd, log(Label(i, 3)))
    pic.CurrentX = xx + dd
    pic.CurrentY = log(Label(i, 3)) - dh
    If Label(i, 3) < 10 Then
      pic.Print " " + Format(Round(Label(i, 3), 1))
    Else
      pic.Print CInt(Label(i, 3))
    End If
  Next i
End Sub

Public Sub show(ByRef pic As PictureBox, plotMode As Integer)
  Dim i As Integer, r As Single
  Dim y0 As Single, yy As Single

  mode = plotMode   ' 0-Screen, 1-Copy, 2-Print
  calc
  init pic
  If mode = 2 Then drawTitle pic, plotName
  drawAxis pic

  'X scale
  setScale pic, 0, 10, 0, 10
  Xscale pic
  'left Y scale
  setScale pic, xmin, xmax, ymin, ymax
  Lscale pic, ymin, ymax
  'grains
  r = pic.ScaleWidth / 80
  pic.FillStyle = 1
  For i = 1 To Num
    pic.Circle (data(i, 1), data(i, 2)), r, colGrain
  Next i
  'right Y scale
  y0 = log(Label(1, 3))
  yy = log(Label(11, 3))
  setScale pic, xmin, xmax, y0, yy
  Rscale pic, y0, yy
  'peaks
  For i = 1 To PkNum
      pic.Line (xmin, log(Label(6, 3)))-(xmax, log(PkAge(i))), colPeak
  Next i
End Sub

'------------------------ plot undepended methods --------------------------------

Private Sub init(pic As PictureBox)
  Dim hh As Single

  pic.Cls
  pic.Font = Font
  If mode = 0 Then
    pic.Font.Size = screenFontSize
  Else
    pic.Font.Size = printFontSize
  End If
  pic.FillStyle = 0
  pic.BackColor = colBack
  pic.FillColor = colFill
  pic.ScaleMode = 0    'custom
  pic.Scale (0, pic.Height)-(pic.Width, 0)
  hh = Abs(pic.ScaleHeight)
  hFont = Abs(pic.TextHeight("0"))
  If mode = 0 Then                                  'screen
    psize = hh - cScreen * hFont
  ElseIf mode = 1 Then                              'copy
    psize = hh - cCopy * hFont
  ElseIf mode = 2 Then                              'print
    psize = hh - cPrint * hFont
  End If
  dh2 = (hh - psize) / hh
  dw2 = (pic.ScaleWidth - psize) / pic.ScaleWidth
End Sub
  
Private Function getY0(pic As PictureBox)
  Dim hh As Single

  hh = Abs(pic.ScaleHeight)
  If mode = 0 Then
    getY0 = pic.ScaleTop - hh + (hh * dh2) * (1 - cTopScreen)
  ElseIf mode = 1 Then
    getY0 = pic.ScaleTop - hh + (hh * dh2) * (1 - cTopCopy)
  ElseIf mode = 2 Then
    getY0 = pic.ScaleTop - hh + (hh * dh2) * (1 - cTopPrint)
  End If
End Function

Private Function getYY(pic As PictureBox)
  Dim hh As Single

  hh = Abs(pic.ScaleHeight)
  If mode = 0 Then
    getYY = pic.ScaleTop - (hh * dh2) * cTopScreen
  ElseIf mode = 1 Then
    getYY = pic.ScaleTop - (hh * dh2) * cTopCopy
  ElseIf mode = 2 Then
    getYY = pic.ScaleTop - (hh * dh2) * cTopPrint
  End If
End Function

Private Function getX0(pic As PictureBox)
  getX0 = pic.ScaleLeft + (pic.ScaleWidth * dw2) / 2
End Function

Private Function getXX(pic As PictureBox)
  getXX = pic.ScaleLeft + pic.ScaleWidth - (pic.ScaleWidth * dw2) / 2
End Function

Private Sub drawTitle(pic As PictureBox, name As String)
  pic.Font.Size = 1.1 * printFontSize
  pic.Font.Bold = True
  pic.CurrentX = (pic.ScaleWidth - pic.TextWidth(name)) / 2
  pic.CurrentY = pic.ScaleTop
  pic.Print name
  pic.Font.Size = printFontSize
  pic.Font.Bold = False
End Sub

Private Sub drawAxis(pic As PictureBox)
  Dim yText As New Rotator

  pic.Line (getX0(pic), getY0(pic))-(getXX(pic), getYY(pic)), , B
  If mode <> 1 Then
    'X
    pic.CurrentX = pic.ScaleLeft + (pic.ScaleWidth - pic.TextWidth(xLabel)) / 2
    pic.CurrentY = 1.2 * hFont
    pic.Print xLabel
    ' Left
    Set yText.Device = pic
    yText.Angle = 90
    pic.CurrentY = getY0(pic) + (psize - pic.TextWidth(yLeft)) / 2
    pic.CurrentX = 0.2 * hFont
    yText.Label yLeft
    ' Right
    Set yText.Device = pic
    yText.Angle = 90
    pic.CurrentY = getY0(pic) + (psize - pic.TextWidth(yRight)) / 2
    pic.CurrentX = pic.ScaleWidth - 1.2 * hFont
    yText.Label yRight
  End If
End Sub

Private Sub setScale(pic As PictureBox, ByVal x0 As Single, ByVal xx As Single, ByVal y0 As Single, ByVal yy As Single)
  Dim dx As Single, dy As Single

  dx = (xx - x0) * dw2 / (1 - dw2)
  dy = (yy - y0) * dh2 / (1 - dh2)
  If mode = 0 Then
    pic.Scale (x0 - dx / 2, yy + cTopScreen * dy)-(xx + dx / 2, y0 - (1 - cTopScreen) * dy)
  ElseIf mode = 1 Then
    pic.Scale (x0 - dx / 2, yy + cTopCopy * dy)-(xx + dx / 2, y0 - (1 - cTopCopy) * dy)
  ElseIf mode = 2 Then
    pic.Scale (x0 - dx / 2, yy + cTopPrint * dy)-(xx + dx / 2, y0 - (1 - cTopPrint) * dy)
  End If
  hFont = Abs(pic.TextHeight("0"))
End Sub

