Attribute VB_Name = "ZetaAge"
'Converted from ZetaAge
Option Explicit

Private Const Errn = 0.000001  'Max error for confidence interval calculation

Public Const PFmax = 50                      'max potential peaks
Public pPeaks(0 To PFmax, 0 To 3) As Single  'potential peaks
Public pNum As Integer                       'number of potential peaks

Public Sub CalcPP()
  Dim i As Integer
  If Num > 0 Then
    toLogLine "# Potential peaks:"
    pNum = FindPeaks(1, 1)    ' first derivative list
    pNum = pNum + FindPeaks(2, 1)   ' second derivative list
    pNum = peaksCorrection()
    toLogLine "         Age       PD    Est.N"
    For i = 1 To pNum
      pPeaks(i, 0) = 1
      toLogLine nFormat(i, 0, 3) + nFormat(pPeaks(i, 1), 2, 9) + nFormat(pPeaks(i, 2), 2, 9) + nFormat(pPeaks(i, 3), 2, 9)
    Next i
    toLogLine "  ----------------------------"
    toLogLine "  Pool Age: " + nFormat(pPeaks(0, 1), 2, 9) + " Ma"
    toLogLine ""
  End If
End Sub

Private Function peaksCorrection() As Integer
  Dim i As Integer
  Dim n As Integer
  Dim k As Integer
  Dim Done As Integer

  n = 0
  k = 0
  Done = 0
  For i = 1 To pNum
    If CInt(pPeaks(i, 1)) <= CInt(AgeMin) Then
      If k = 0 Then
        k = 1
        n = n + 1
      End If
      If pPeaks(i, 1) < AgeMin Then
        pPeaks(n, 1) = AgeMin
      Else
        pPeaks(n, 1) = pPeaks(i, 1)
      End If
    ElseIf CInt(pPeaks(i, 1)) < CInt(AgeMax) Then
      n = n + 1
      pPeaks(n, 1) = pPeaks(i, 1)
    Else
      n = n + 1
      If pPeaks(i, 1) < AgeMax Then pPeaks(n, 1) = pPeaks(i, 1) Else pPeaks(n, 1) = AgeMax
      Done = 1
    End If
    pPeaks(n, 2) = pPeaks(i, 2)
    pPeaks(n, 3) = pPeaks(i, 3)
    If Done = 1 Then Exit For
  Next i
  peaksCorrection = n
End Function

Private Function FindPeaks(dmode As Integer, mode As Integer) As Integer
  Dim pNum As Integer
  Dim amin, amax
  Dim SumNs As Single
  Dim SumNi As Single
  Dim i As Integer
  Dim PoolAge As Single, PoolAgeL As Single, PoolAgeU As Single

  If dmode = 1 Then
    pNum = PkFind1(amin, amax)
  ElseIf dmode = 2 Then
    pNum = PkFind2(amin, amax)
  Else
    pNum = PkFind2(amin, amax)
    If pNum > PFmax Then pNum = PkFind1(amin, amax)
    If pNum > PFmax Then pNum = PFmax
  End If

  '... Calculate age and uncertainties for pooled age
  SumNs = 0
  SumNi = 0
  For i = 1 To Num
    SumNs = SumNs + Ns(i)
    SumNi = SumNi + Ni(i)
  Next i
  Call FTeqn(SumNs, SumNi, RhoD0, RERhoD0, Zeta0, REZeta0, mode, PoolAge, PoolAgeL, PoolAgeU)
  pPeaks(0, 1) = PoolAge
  FindPeaks = pNum
End Function

Private Function PkFind1(AgeMin, AgeMax) As Integer
  'output: AgeMin, AgeMax
  ' return number of peaks
  '========================================================================
  '... Finds peaks in the probability density distribution calculated from
  ' information stored in the following shared variables:
  ' FTage(), Num%:     fission track grain ages and number of grain ages
  ' Zgrain(), Zerr():  Z and SE(Z) for the grains
  ' Zerr() should be in units of 1*SE and should not include uncertainties
  ' for monitor track density and zeta factor.

  Dim ZUp(1 To maxGrains) As Single
  Const Zacc = 0.003    '... specified accuracy for Z at peak
  Dim PkNum As Integer
  Dim Zmin As Single, Zmax As Single
  Dim dZ0 As Single, dz As Single
  Dim Flw As Single
  Dim Dum As Single
  Dim Fup As Single
  Dim Zi As Single
  Dim j As Integer, i As Integer
  '========================================================================
  '... Find min and max
  Zmin = Zgrain(1)
  Zmax = Zgrain(1)
  For j = 2 To Num
    If Zgrain(j) > Zmax Then Zmax = Zgrain(j)
    If Zgrain(j) < Zmin Then Zmin = Zgrain(j)
  Next j
  AgeMax = TaufromZ(Zmax)
  AgeMin = TaufromZ(Zmin)

  '... To catch the tails of the distribution, the range for Z is set
  ' at 2 standard errors below and above the minimum and maximum Zgrain.
  Zmax = Zmax + 2 * QMeanZerr
  Zmin = Zmin - 2 * QMeanZerr
  '... C0 is defined as a constant in the main program.
  '... The uncertainty for each grain age is the grain-only standard error.
  PkNum = 0
  dZ0 = (Zmax - Zmin) / Num
  Call KernelPD(Zmin, 1, Flw, Dum)
  For Zi = (Zmin + dZ0) To Zmax Step dZ0
    Call KernelPD(Zi, 1, Fup, Dum)
    '... Maximum in the density function is indicated by the 1rd derivative
    ' changing from positive to negative with increasing Zi.
    If Flw > 0 And Fup <= 0 Then
      PkNum = PkNum + 1
'      ZLw(PkNum) = Zi - dZ
      ZUp(PkNum) = Zi
    End If
    Flw = Fup
    If PkNum = Num Then Exit For
  Next Zi

  Dim UpZ As Single
  Dim LwZ As Single
  Dim f As Single
  Dim Fmid As Single
  Dim RTbis As Single
  Dim MidZ As Single
  Dim PkDen As Single
  Dim PkArea As Single
  Dim PkAge As Single
  Dim n As Integer
  n = 0
  For i = 1 To PkNum
    UpZ = ZUp(i)
    LwZ = ZUp(i) - dZ0
    Call KernelPD(LwZ, 1, f, Dum)
    Call KernelPD(UpZ, 1, Fmid, Dum)
    If f <= 0 Or Fmid >= 0 Then                   '!!!
       MsgBox "Incorrect brackets for bisection"
       Stop
    End If
    RTbis = UpZ
    dz = LwZ - UpZ
    For j = 1 To 40
      dz = dz * 0.5
      MidZ = RTbis + dz
      Call KernelPD(MidZ, 1, Fmid, Dum)
      If Fmid <= 0 Then RTbis = MidZ
      If (Abs(dz) < Zacc) Or (Fmid = 0) Then Exit For
    Next j
    If j = 40 Then MsgBox "Too many bisections"     '!!!

    '... Calculate peak density and peak area
    ' Peak area = peak density * std dev *Sqr(2*Pi)
    Call KernelPD(RTbis, 0, PkDen, Dum)
    PkArea = PkDen * QMeanZerr / C0
    PkDen = 0.1 * PkDen

    '... Convert from the Z scale to the Tau scale.
    PkAge = TaufromZ(RTbis)

    '... Record location of peak
    If PkNum <= PFmax Then
      n = n + 1
      pPeaks(n, 1) = PkAge
      pPeaks(n, 2) = PkDen
      pPeaks(n, 3) = PkArea
    End If
  Next i
  PkFind1 = n
End Function

Private Function PkFind2(AgeMin, AgeMax) As Integer
  'output: AgeMin, AgeMax, PkNum(global)
  'return number of peaks
  '========================================================================
  '... Finds peaks in the probability density distribution calculated from
  ' information stored in the following shared variables:
  ' FTage(), Num%:     fission track grain ages and number of grain ages
  ' Zgrain(), Zerr():  Z and SE(Z) for the grains
  ' Zerr() should be in units of 1*SE and should not include uncertainties
  ' for monitor track density and zeta factor.
  '========================================================================
  Dim ZUp(1 To maxGrains) As Single
  Const Zacc = 0.003    '... specified accuracy for Z at peak
  Dim PkNum As Integer
  Dim Zmin As Single, Zmax As Single
  Dim dZ0 As Single, dz As Single
  Dim Flw As Single
  Dim Dum As Single
  Dim Fup As Single
  Dim Zi As Single
  Dim j As Integer, i As Integer

  '... Find min and max
  Zmin = Zgrain(1)
  Zmax = Zgrain(1)
  For j = 2 To Num
    If Zgrain(j) > Zmax Then Zmax = Zgrain(j)
    If Zgrain(j) < Zmin Then Zmin = Zgrain(j)
  Next j
  AgeMax = TaufromZ(Zmax)
  AgeMin = TaufromZ(Zmin)

  '... To catch the tails of the distribution, the range for Z is set
  ' at 2 standard errors below and above the minimum and maximum Zgrain.
  Zmax = Zmax + 2 * QMeanZerr
  Zmin = Zmin - 2 * QMeanZerr

  PkNum = 0
  dZ0 = (Zmax - Zmin) / Num
  Call KernelPD(Zmin, 3, Flw, Dum)
  For Zi = (Zmin + dZ0) To Zmax Step dZ0
    Call KernelPD(Zi, 3, Fup, Dum)
    '... Minimum in the 2nd derivative is indicated by the 3rd derivative
    ' changing from negative to positive with increasing Zi.
    If Flw < 0 And Fup > 0 Then
      PkNum = PkNum + 1
      ZUp(PkNum) = Zi
    End If
    Flw = Fup
    If PkNum = Num Then Exit For
  Next Zi

  Dim UpZ As Single
  Dim LwZ As Single
  Dim f As Single
  Dim Fmid As Single
  Dim RTbis As Single
  Dim MidZ As Single
  Dim PkDen As Single
  Dim PkArea As Single
  Dim PkAge As Single
  Dim n As Integer

  n = 0
  For i% = 1 To PkNum%
    UpZ = ZUp(i%)
    LwZ = UpZ - dZ0
    Call KernelPD(LwZ, 3, f, Dum)
    Call KernelPD(UpZ, 3, Fmid, Dum)
    If f >= 0 Or Fmid <= 0 Then               '!!!
      MsgBox "Incorrect brackets for bisection"
      Stop
    End If
    RTbis = LwZ
    dz = UpZ - LwZ
    For j = 1 To 40
      dz = dz * 0.5
      MidZ = RTbis + dz
      Call KernelPD(MidZ, 3, Fmid, Dum)
      If Fmid <= 0 Then RTbis = MidZ
      If (Abs(dz) < Zacc) Or (Fmid = 0) Then Exit For
    Next j
    If j% = 40 Then MsgBox "Too many bisections"     '!!!

    '... Calculate peak density and peak area
    ' Peak area = peak density * std dev *Sqr(2*Pi)
    Call KernelPD(RTbis, 0, PkDen, Dum)
    PkArea = PkDen * QMeanZerr / C0
    PkDen = 0.1 * PkDen

    '... Convert from the Z scale to the Tau scale.
    PkAge = TaufromZ(RTbis)

    '... Record location of peak
    If PkNum <= PFmax Then
      n = n + 1
      pPeaks(n, 1) = PkAge
      pPeaks(n, 2) = PkDen
      pPeaks(n, 3) = PkArea
    End If
  Next i
  PkFind2 = n
End Function

Public Sub KernelPD(Zi As Single, Order As Integer, pd As Single, SEPd As Single)
  ' input: Zi, Order
  ' output PD, SEPd
  '============================================================================
  '... Calculates the probability density function and its derivatives
  ' at the specified value of Zi. When Order = -1, the routine returns the
  ' estimated probability density and standard error. Otherwise SEPD == 0.
  ' Shared variables: Zgrain(), Zerr(), Num%
  ' Constant: KernelFactor, which is set in main program

  Dim SSPD As Single
  Dim dev As Single
  Dim dev2 As Single
  Dim S As Single
  Dim S2 As Single, S3 As Single, S5 As Single, S7 As Single
  Dim Term As Single
  Dim ExpTerm As Single
  Dim i As Integer

  pd = 0
  SSPD = 0
  For i = 1 To Num
    dev = Zi - Zgrain(i)
    S = KernelFactor * Zerr(i)
    If Abs(dev) < 5 * S Then
        dev2 = dev * dev
        S2 = S * S
        ExpTerm = Exp(-dev2 / (2 * S2))
        Select Case Order
            Case -1
                Term = (C0 / S) * ExpTerm
                pd = pd + Term
                SSPD = SSPD + Term * Term
            Case 0
                pd = pd + (C0 / S) * ExpTerm
            Case 1
                S3 = S * S2
                pd = pd - (C0 * dev / S3) * ExpTerm
            Case 2
                S5 = S2 * S2 * S
                pd = pd + (C0 * (dev2 - S2) / S5) * ExpTerm
            Case 3
                S7 = S2 * S2 * S2 * S
                pd = pd + (C0 * dev * (-dev2 + 3 * S2) / S7) * ExpTerm
        End Select
    End If
  Next i
  If Order = -1 And Num > 1 Then
    '... SE of estimated probability density; formula is after
    ' eqn. 3.7 in Silverman (1986, p. 36). Note that the formula
    ' accounts for the fact that PD has units of number of grains
    ' per Num% grains per unit Z.
    SEPd = Sqr((1! * Num * SSPD - pd * pd) / (1! * Num - 1))
  Else
    SEPd = 0
  End If
End Sub

Public Sub FTeqn(Ns As Single, Ni As Single, RhoD As Single, RERhoD As Single, Zeta As Single, REZeta As Single, CIOpt As Integer, age As Single, AgeL As Single, AgeU As Single)
  '============================================================================
  '... Calculates FT age and confidence interval and includes uncertainties
  ' for spontaneous, induced, and monitor track densities, and the zeta factor.
  ' The FT age is calculated using the Zeta method equation of Hurford and
  ' Green (Isotope Geoscience, 1983) and the formulation of Galbraith
  ' (Nuclear Tracks, 1990). Confidence intervals are determined using
  ' the method outlined in Bardsley (Mathematical  Geology, 1983) and Galbraith
  ' (Mathematical Geology, 1984; ibid, 1990). When Ns or Ni is small, a median
  ' estimate and 95% confidence interval are calculated for the parameter
  ' (Ns*Nd/Ni) using Sneyd's algorithm (Computers and Geosciences, 1984).
  ' Constants LamdaD, and GF are set in main program. The geometric
  ' correction for the fluence monitor is explictly included as GF, so that the
  ' induced tracks (Ni) should be given as the actual number of counted tracks.
  '
  'INPUT:  Ns, Ni = spontaneous and induced tracks.
  '        RhoD, RErhodD = monitor track density and relative standard error.
  '        Zeta, REzeta = zeta factor and relative standard error.
  '        CIOpt% = option for confidence interval: 1 = 68%, 2 = 95%
  'OUTPUT: Mean age and 95% confidence interval in Ma.
  '============================================================================
  Dim Zfactor As Single
  Dim Alpha As Single
  Dim SE As Single
  Dim ExpZm As Single, ExpZ1 As Single, ExpZ2 As Single
  Dim ND As Single, N1 As Long, N2 As Long, N3 As Long
  Dim CtArea As Single
  Dim ExpZAm As Single, ExpZA1 As Single, ExpZA2 As Single

  '... Select confidence interval
  Select Case CIOpt
    Case 1  '... 68% confidence interval
      Zfactor = 1
      Alpha = (1 - 0.68) / 2
    Case 2  '... 95% confidence interval
      Zfactor = 1.96
      Alpha = (1 - 0.95) / 2
  End Select

  '... Test for applicability of normal-distribution approximation.
  ' My experience is that the normal approximation can be used with the
  ' Z transform when SE(Z) < 0.15.
  SE = Sqr(1 / (Ns + 0.5) + 1 / (Ni + 0.5) + RERhoD * RERhoD + REZeta * REZeta)
  If SE < 0.15 Then
    '... Use the normal distribution with the Z transform to find
    ' confidence limits.
    ExpZm = LamdaD * Zeta * GF * RhoD * (Ns + 0.05) / (Ni + 0.5)
    ExpZ1 = ExpZm * Exp(-Zfactor * SE)
    ExpZ2 = ExpZm * Exp(Zfactor * SE)
  Else
    '... Find exact confidence limits for cases where Ns or Ni is small.
    ' Zeta uncertainties are assumed to be relatively small and therefore
    ' are not included in this calculation.
    '... Estimate equivalent track count for effective monitor track density
    If RERhoD <> 0 Then
        ND = 1 / (RERhoD * RERhoD)
    Else
        ND = 10000
    End If
    CtArea = ND / RhoD
    ND = CLng(ND)
    N1 = CLng(Min(Ns, ND))
    N2 = CLng(Max(Ns, ND) - N1)
    N3 = CLng(N1 + Ni)
    '... Calculate mean and initial estimates of confidence limits
    SE = (Ns / ((Ns + 0.5) * (Ns + 0.5))) + (Ni / ((Ni + 0.5) * (Ni + 0.5)))
    SE = Sqr(SE + RERhoD * RERhoD)
    ExpZAm = (ND + 0.5) * (Ns + 0.5) / (Ni + 0.5)
    ExpZA1 = ExpZAm * Exp(-Zfactor * SE)
    ExpZA2 = ExpZAm * Exp(Zfactor * SE)
    '... Solve for exact confidence limits
    Call Solve(Alpha, N1, N2, N3, ExpZA2, ExpZ2, 1)
    Call Solve(0.5, N1, N2, N3, ExpZAm, ExpZm, 2)
    N1 = Max(0!, (N1 - 1!))
    Call Solve(1! - Alpha, N1, N2, N3, ExpZA1, ExpZ1, 1)
    ExpZ1 = LamdaD * Zeta * GF * ExpZ1 / CtArea
    ExpZm = LamdaD * Zeta * GF * ExpZm / CtArea
    ExpZ2 = LamdaD * Zeta * GF * ExpZ2 / CtArea
  End If
  AgeL = log(1 + ExpZ1) / (LamdaD * 1000000!)
  age = log(1 + ExpZm) / (LamdaD * 1000000!)
  AgeU = log(1 + ExpZ2) / (LamdaD * 1000000!)
End Sub

Private Sub Solve(a As Single, N1 As Long, N2 As Long, N3 As Long, ThetaA As Single, Theta As Single, mode As Integer)
  '============================================================================
  '... This routine solves: FN(Theta) = Log(A)
  ' where ThetaA is the initial guess.
  '============================================================================
  Dim Al As Single
  Dim SStep As Single
  Dim Theta1 As Single, Theta2 As Single
  Dim S2 As Single
  Dim Ptest As Single

  Al = log(1 / a)
  '... Find an interval containing the solution
  Theta = ThetaA
  SStep = Sign((Theta / 10), (Xzdy(Theta, N1, N2, N3, mode) + Al))

  Theta1 = Theta + SStep
  Theta2 = Theta
  S2 = Sign(1, SStep)
  Ptest = Xzdy(Theta1, N1, N2, N3, mode) + Al
  Do While S2 = Sign(1, Ptest)
    Theta1 = Theta1 + SStep
    If Theta1 <= 0 Then
      Theta1 = 0.5 * (Theta1 - SStep)
      SStep = -0.2 * (Theta1 - SStep)
    End If
    Ptest = Xzdy(Theta1, N1, N2, N3, mode) + Al
  Loop

  '... Carry out bisection
  Do While Abs(Theta1 - Theta2) > (Max(1!, Theta1) * Errn)
    Theta = 0.5 * (Theta1 + Theta2)
    If Sign(1, Xzdy(Theta, N1, N2, N3, mode) + Al) = S2 Then
      Theta2 = Theta
    Else
      Theta1 = Theta
    End If
  Loop
End Sub

Public Function getPF(ByVal Chi1 As Single, ByVal DF1 As Integer, ByVal Chi2 As Single, ByVal df2 As Integer) As Single
  Dim f As Single, x As Single
  Dim Prob As Single
  f = (Chi1 - Chi2) / (Chi2 / df2)
  DF1 = DF1 - df2
  x = df2 / (df2 + DF1 * f)
  Prob = Betai(df2 / 2, DF1 / 2, x)
  getPF = Prob * 100
End Function

Private Function Sign(a As Single, b As Single) As Single
  '============================================================================
  '... Replaces non-standard sign function in Qbasic
  '============================================================================
  If b <> 0 Then Sign = Abs(a) * (b / Abs(b)) Else Sign = Abs(a)
End Function

Private Function Vxzdy(x As Single, N1 As Long, N2 As Long, N3 As Long, Mi As Long, Mend As Long) As Double
  '============================================================================
  '... Calculates the sum from R = 0 to Mend of:
  '         X^R / (R! * (N2%+R)! * (N3%-R)!)
  ' which is then divided by the R = Mi term.
  '============================================================================
  Dim S As Double
  Dim Term As Double
  Dim i As Long

  S = 1
  Term = 1
  i = Mi + 1
  Do While (Term > Errn) And (i <= Mend)
    Term = Term * (x * (1! * N3 - i + 1) / (1! * i * (N2 + i)))
    S = S + Term
    i = i + 1
  Loop
  Term = 1
  i = Mi
  Do While (Term > Errn) And (i > 0)
    Term = Term / (x * (N3 - i + 1!) / (1! * i * (N2 + i)))
    S = S + Term
    i = i - 1
  Loop
  Vxzdy = S
End Function

Private Static Function Xzdy(Theta As Single, N1 As Long, N2 As Long, N3 As Long, mode As Integer) As Single
  '============================================================================
  '... For mode = 1, this routine calculates:
  '       Log(Q(X, Theta) / Q(X + Y, Theta))
  '... For mode = 2, this routine calculates:
  '       Log(average of XZDY(Theta) for N1% and N1%-1)
  ' Note that: N1% = Ns, N2% = Nd-Ns, N3% = Ns+Ni
  '============================================================================
  Dim Ml As Long, Mu As Long

  Dim x As Single
  Dim Mln As Long, Mun As Long
  Dim Fixx As Single
  Dim R1 As Single, R2 As Single, R3 As Single, R4 As Single, Rd As Single
  Dim LogVup As Single, LogVdwn As Single
  Dim NN1 As Long
  Dim f As Single

  Select Case mode
    Case 1
      '... Calculates orders ML and MU of max terms in Q1(X,TH) and Q(X+Y,TH)
      x = Sqr(((N2 + Theta) * (N2 + Theta)) + (4 * Theta * (1 + N3)))
      x = 0.5 * (x - N2 - Theta)
      Mln = CInt(x)
      If Mln >= N3 Then Mln = 0
      If N1 < Mln Then Mun = N1 Else Mun = Mln

      '... If Mu or Ml have been changed since last call then recalculate:
      '  FIX = Log(Mu - Coeff in Q(X,Theta)) - Log(Ml - Coeff in Q(X+Y, Theta))
      If Mln <> Ml Or Mun <> Mu Then
        Ml = Mln
        Mu = Mun
        Fixx = 0
        If Ml <> Mu Then
          R1 = 1! * N2 + Ml
          R2 = 1! * N2 + Mu
          R3 = 1! * N3 - Ml
          R4 = 1! * N3 - Mu
          Rd = 1! * Ml - Mu
          If Mu = 0 Then
            Fixx = 0.918939 + (Ml + 0.5) * log(Ml)
          Else
            Fixx = Rd * log(Ml) + (Mu + 0.5) * log(1! * Ml / Mu)
          End If
          Fixx = Fixx - Rd + Rd * log(R1 / R4) + (R3 + 0.5) * log(R1 * R3 / (R2 * R4)) + (R2 - R3) * log(R1 / R2)
          Fixx = Fixx + Facres(Ml) + Facres(Ml + N2) + Facres(N3 - Ml) - Facres(Mu + N2) - Facres(N3 - Mu)
          If Mu > 0 Then Fixx = Fixx - Facres(Mu)
        End If
      End If
      LogVup = log(Vxzdy(Theta, N1, N2, N3, Mu, N1))
      LogVdwn = log(Vxzdy(Theta, N1, N2, N3, Ml, N3))
      Xzdy = (1! * Mu - Ml) * log(Theta) + Fixx + LogVup - LogVdwn
    Case 2
      f = Exp(Xzdy(Theta, N1, N2, N3, 1))
      NN1 = Max(0!, N1 - 1!)
      f = 0.5 * (f + Exp(Xzdy(Theta, NN1, N2, N3, 1)))
      Xzdy = log(f)
  End Select
End Function

Private Function Facres(n As Long) As Single
  '============================================================================
  '... Calculates Log(N) - (N+0.5) * Log(N) + N - Log(2*Pi), which provides
  ' the correction to Stirling's formula for confidence interval calculation.
  '============================================================================
  Dim f As Single
  Dim i As Integer
  Dim PowerN3 As Single, PowerN5 As Single

  f = 0
  If n <= 4 Then
    For i = 2 To n
      f = f + log(i)
    Next i
    f = f - (n + 0.5) * log(n) + n - 0.918939
  Else
    PowerN3 = 1! * n * n * n
    PowerN5 = PowerN3 * n * n
    f = (0.08333333 / n) - (0.002777777 / PowerN3) + (0.0007936508 / PowerN5)
  End If
  Facres = f
End Function

Private Function Betacf(a As Single, b As Single, x As Single) As Single
  '===========================================================================
  '... Continued fraction for incomplete beta function, used by Betai
  '===========================================================================
  Dim ITMAX As Integer
  Dim EPS As Single
  Dim AM As Single, BM As Single, AZ As Single, BZ As Single
  Dim QAB As Single, QAP As Single, QAM As Single
  Dim M As Integer
  Dim EM As Single, TEM As Single, d As Single
  Dim AP As Single, BP As Single, APP As Single, BPP As Single, AOLD As Single

  ITMAX% = 100
  EPS = 0.0000003
  AM = 1!
  BM = 1!
  AZ = 1!
  QAB = a + b
  QAP = a + 1!
  QAM = a - 1!
  BZ = 1! - QAB * x / QAP
  For M = 1 To ITMAX
    EM = M
    TEM = EM + EM
    d = EM * (b - M%) * x / ((QAM + TEM) * (a + TEM))
    AP = AZ + d * AM
    BP = BZ + d * BM
    d = -(a + EM) * (QAB + EM) * x / ((a + TEM) * (QAP + TEM))
    APP = AP + d * AZ
    BPP = BP + d * BZ
    AOLD = AZ
    AM = AP / BPP
    BM = BP / BPP
    AZ = APP / BPP
    BZ = 1!
    If Abs(AZ - AOLD) < EPS * Abs(AZ) Then Exit For
  Next M
  If Abs(AZ - AOLD) >= EPS * Abs(AZ) Then                  '!!!
    MsgBox "A or B too big, or ITMAX too small (func.: Betacf)"
  Else
    Betacf = AZ
  End If
End Function

Private Function Betai(a As Single, b As Single, x As Single) As Single
  '===========================================================================
  '... Returns the incomplete beta function: Ix(a,b)
  '===========================================================================
  Dim BT As Single

  If x < 0! Or x > 1! Then            '!!!
    MsgBox "Bad argument X in BETAI"
    Exit Function
  End If
  If x = 0! Or x = 1! Then
    BT = 0!
  Else
    BT = Exp(GammLn(a + b) - GammLn(a) - GammLn(b) + a * log(x) + b * log(1! - x))
  End If
  If x < (a + 1!) / (a + b + 2!) Then
    Betai = BT * Betacf(a, b, x) / a
  Else
    Betai = 1! - BT * Betacf(b, a, 1! - x) / b
  End If
End Function

Private Function Max(a As Single, b As Single) As Single
  If a > b Then Max = a Else Max = b
End Function

Private Function Min(a As Single, b As Single) As Single
  If a < b Then Min = a Else Min = b
End Function

