Attribute VB_Name = "bfitMain"
Option Explicit

Public Const MaxFiles = 20
Public Const maxGrains = 1000
Public Const maxPeaks = 10
Public msgMaxGrains As String
Public msgMaxFiles As String

Public LogLike As Single
Public RChiSq As Single
Public ChiSq As Single
Public Cnumb As Single

Public PkTheta0(1 To maxPeaks) As Single
Public PkFrac0(1 To maxPeaks) As Single

'BINOMFIT objects:
Public data As New dataFiles
Public PDplot As New PD_plot
Public radPlot As New RadialPlot
Public grPlot As New grainsPlot
Public PFplot As New PF_plot
'

Sub Main()
  APP.title = "BinomFit for Windows"
  progname = "BinomFit"
  msgMaxFiles = "Cannot add file !" + vbCrLf + "(maximum files = " + Str(MaxFiles) + ")"
  binom.Top = (Screen.Height - binom.Height) / 2
  binom.Left = (Screen.Width - binom.Width) / 2
  binom.show
End Sub

Public Sub PkCalc()
  '============================================================================
  ' PkAge() = age of selected peaks
  ' PkTheta() = theta ratio for each peak
  ' Pkfrac() = fractional proportion of the total distribution of each peak
  '        where Sum[Pkfrac()] = 1
  ' Num (shared) = number of grain ages or peaks in the input dataset
  ' PkNum (shared) = total number of peaks
  ' Zgrain(), ZErr() (shared) = Z and SE(Z) for grains
  ' AgeMin, AgeMax (shared) = age range for grain-age distribution
  '============================================================================
  Dim i As Integer
  Dim Zpk As Single, PkDen As Single
  '... Make a guess of proportional size of each peak
  '!!! Me.MousePointer = 11
  For i = 1 To PkNum
    PkTheta(i) = Theta(PkAge(i))
    '... Calculate probability density at the peak.
    Zpk = ZfromTau(PkAge(i))
    PkDen = SumPD(Zpk, 0)
    '... Fraction of total grains = peak density * std dev *Sqr(2*Pi)/ Num
    ' Use a SD(z) = 0.15 to estimate the proportion of each peak
    ' relative to the total distribution.
    PkFrac(i) = PkDen * 0.15 / (C0 * Num)
      '... Set a lower limit for PkFrac(I%) to avoid division-by-zero problems.
    If PkFrac(i) < 0.001 Then PkFrac(i) = 0.001
Next i

  For i = 1 To PkNum
    PkTheta0(i) = PkTheta(i)
    PkFrac0(i) = PkFrac(i)
  Next i

  LogLike = EM(PkTheta(), PkFrac())
  Cnumb = StdErr()
  For i = 1 To PkNum
    PkSDz(i) = PkWidthFactor * PkSDz(i)
  Next i
  ChiSq = Chi2(PkTheta(), PkFrac())
  RChiSq = ChiSq / (Num - (2 * PkNum - 1))
  '!!! Me.MousePointer = 0
End Sub

Function Theta(age As Single) As Single
  '============================================================================
  '... Given a FT age, this routine calculates Theta = Ns /(Ns+Ni).
  ' Age = FT age in Ma
  ' Zeta0 (shared) = zeta factor
  ' RhoD0 (shared) = monitor track density used for this dataset
  ' LamdaD (shared) = total decay constant for 238U (yr^-1)
  ' GF (shared) = geometry factor
  '============================================================================
  Dim ExpTL As Single
  ExpTL = Exp(age * 1000000! * LamdaD)
  Theta = (ExpTL - 1) / (LamdaD * Zeta0 * GF * RhoD0 + ExpTL - 1)
End Function

Function ZfromTau(Tau As Single) As Single
  '============================================================================
  '... Converts to Z from age in Myr.
  ' LamdaD is a constant set in the main program.
  '============================================================================
  ZfromTau = log(Exp(Tau * 1000000! * LamdaD) - 1)
End Function

Function Chi2(PkTheta() As Single, PkFrac() As Single) As Single
  '============================================================================
  '... Calculates Chi^2 parameter for the binomial model of Galbraith
  ' and Green. PkTheta() contains the theta ratios for the best-fit peaks, and
  ' Pkfrac() contains the proportions of the total distribution for the peaks.
  ' Num% (shared) = number of dated grains
  ' PkNum% (shared) = number of peaks to be found
  ' Ns() (shared) = number of spontaneous tracks counted per grain
  ' Nt() (shared) = total number of tracks counted per grain
  ' Chi2 = function is returned with Chi^2 value
  '============================================================================
  '... Find mean and sum of squares for solution vector
  Dim Mu As Single, ss As Single
  Dim Factor1 As Single, Factor2 As Single
  Dim nnt As Single, dev As Single, Var As Single, C2 As Single
  Dim i As Integer

  Mu = 0
  ss = 0
  For i = 1 To PkNum
    Mu = Mu + PkFrac(i) * PkTheta(i)
    ss = ss + PkFrac(i) * PkTheta(i) * PkTheta(i)
  Next i

  '... Find Chi2 misfit for this particular solution vector
  Factor1 = Mu * (1 - Mu)
  Factor2 = ss - Mu * Mu
  C2 = 0
  For i = 1 To Num
    nnt = Nt(i)
    '... if Nt(i) = 0, then Ns(i)=0, so contribution to Chi2 is zero.
    If nnt > 0 Then
      dev = Ns(i) - Mu * nnt
      Var = Factor1 * nnt + nnt * (nnt - 1) * Factor2
      C2 = C2 + dev * dev / Var
    End If
  Next i
  Chi2 = C2
End Function

Function SumPD(Zi As Single, Order As Integer) As Single
  '============================================================================
  '... Calculates the probability density function and its derivatives
  ' at the specified value of Zi.
  ' Shared variables: Zgrain(), Zerr(), Num
  '============================================================================
  Dim Dsum As Single, dev As Single, dev2 As Single
  Dim S As Single, S2 As Single, S3 As Single, S5 As Single, S7 As Single
  Dim ExpTerm As Single
  Dim i As Integer
  Dsum = 0
  For i = 1 To Num
    dev = Zi - Zgrain(i)
    S = KernelFactor * Zerr(i)
    If Abs(dev) < 5 * S Then
      dev2 = dev * dev
      S2 = S * S
      ExpTerm = Exp(-dev2 / (2 * S2))
      Select Case Order
      Case 0
        Dsum = Dsum + (C0 / S) * ExpTerm
      Case 1
        S3 = S * S2
        Dsum = Dsum - (C0 * dev / S3) * ExpTerm
      Case 2
        S5 = S2 * S2 * S
        Dsum = Dsum + (C0 * (dev2 - S2) / S5) * ExpTerm
      Case 3
        S7 = S2 * S2 * S2 * S
        Dsum = Dsum + (C0 * dev * (-dev2 + 3 * S2) / S7) * ExpTerm
      End Select
    End If
  Next i
  SumPD = Dsum
End Function

'!!! common
Public Function sRight(ss As String, n As Integer)
  If Len(ss) < n Then
    sRight = Space(n - Len(ss)) + ss
  Else
    sRight = ss
  End If
End Function

Public Function sCenter(ss As String, n As Integer)
  Dim k As Integer, l As Integer

  l = Len(ss)
  k = CInt((n - l) / 2)
  If l < n Then
    sCenter = Space(k) + ss + Space(n - l - k)
  Else
    sCenter = ss
  End If
End Function

Public Function nFormat(ByVal val As Variant, d As Integer, n As Integer, Optional c As String = "") As String
  '... used to ensure d fractional digits and right spaces to make a total length of n characters.
  '... the format functions ensure that on conversion, val is rounded at the correct precision.
  Dim ss As String
  Dim k As Integer
  If d > 0 Then
    ss = Format(val, "0." + String(d, "0"))
  Else
    ss = Format(val, "0")
  End If
  '... used to convert decimal markers from European (,) to US (.) characters
  k = InStr(ss, ",")
  If k > 0 Then ss = Left(ss, k - 1) + "." + Mid(ss, k + 1)
  
  '... optional argument c forces a signed result (+ or -) when set to "+", otherwise only minus is signed
  If c = "+" And val > 0 Then ss = c + ss
  k = Len(ss)
  If n > k Then ss = Space(n - k) + ss
  nFormat = ss
End Function

Static Function Log10(x)
   Log10 = log(x) / log(10#)
End Function

